import numpy as np
from scipy import ndimage

from .batch_filter import BatchFilter


class GrowBoundary(BatchFilter):
    """Grow a boundary between regions in a label array. Does not grow at the
    border of the batch or an optionally provided mask.

    Args:

        labels (:class:`ArrayKey`):

            The array containing labels.

        mask (:class:`ArrayKey`, optional):

            A mask indicating unknown regions. This is to avoid boundaries to
            grow between labelled and unknown regions.

        steps (``int``, optional):

            Number of voxels (not world units!) to grow.

        background (``int``, optional):

            The label to assign to the boundary voxels.

        only_xy (``bool``, optional):

            Do not grow a boundary in the z direction.
    """

    def __init__(self, labels, mask=None, steps=1, background=0, only_xy=False):
        self.labels = labels
        self.mask = mask
        self.steps = steps
        self.background = background
        self.only_xy = only_xy

    def process(self, batch, request):
        gt = batch.arrays[self.labels]
        gt_mask = None if not self.mask else batch.arrays[self.mask]

        if gt_mask is not None:
            # grow only in area where mask and gt are defined
            crop = gt_mask.spec.roi.intersect(gt.spec.roi)

            if crop is None:
                raise RuntimeError(
                    "GT_LABELS %s and GT_MASK %s ROIs don't intersect."
                    % (gt.spec.roi, gt_mask.spec.roi)
                )
            voxel_size = self.spec[self.labels].voxel_size
            crop_in_gt = (
                crop.shift(-gt.spec.roi.offset) / voxel_size
            ).get_bounding_box()
            crop_in_gt_mask = (
                crop.shift(-gt_mask.spec.roi.offset) / voxel_size
            ).get_bounding_box()

            self.__grow(
                gt.data[crop_in_gt], gt_mask.data[crop_in_gt_mask], self.only_xy
            )

        else:
            self.__grow(gt.data, only_xy=self.only_xy)

    def __grow(self, gt, gt_mask=None, only_xy=False):
        if gt_mask is not None:
            assert (
                gt.shape == gt_mask.shape
            ), "GT_LABELS and GT_MASK do not have the same size."

        if only_xy:
            assert len(gt.shape) == 3
            for z in range(gt.shape[0]):
                self.__grow(gt[z], None if gt_mask is None else gt_mask[z])
            return

        # get all foreground voxels by erosion of each component
        foreground = np.zeros(shape=gt.shape, dtype=bool)
        masked = None
        if gt_mask is not None:
            masked = np.equal(gt_mask, 0)
        for label in np.unique(gt):
            if label == self.background:
                continue
            label_mask = gt == label
            # Assume that masked out values are the same as the label we are
            # eroding in this iteration. This ensures that at the boundary to
            # a masked region the value blob is not shrinking.
            if masked is not None:
                label_mask = np.logical_or(label_mask, masked)
            eroded_label_mask = ndimage.binary_erosion(
                label_mask, iterations=self.steps, border_value=1
            )
            foreground = np.logical_or(eroded_label_mask, foreground)

        # label new background
        background = np.logical_not(foreground)
        gt[background] = self.background



import logging
import os
import numpy as np

from gunpowder.array import ArrayKey, Array
from gunpowder.ext import tensorflow as tf
from gunpowder.nodes.generic_train import GenericTrain
from gunpowder.tensorflow.local_server import LocalServer

logger = logging.getLogger(__name__)


class Train(GenericTrain):
    """Tensorflow implementation of :class:`gunpowder.nodes.Train`.

    Args:

        graph (``string``):

            Filename of a tensorflow meta-graph storing the tensorflow graph
            containing an optimizer. A meta-graph file can be created by
            running::

                # create tensorflow graph
                ...

                # store it
                tf.train.export_meta_graph(filename='...')

        optimizer (``string`` or function):

            Either the name of the tensorflow operator performing a training
            iteration, or a function that, given the graph of the meta-graph
            file, adds a custom loss and optimizer.

            If a function is given, it should return a tuple ``(loss,
            optimizer)`` of a tensor and an operator representing the loss and
            the optimizer, respectively. In this case, parameter ``loss``
            should be ``None``.

            Example::

                def add_custom_optimizer(graph):

                    # get the output of your graph
                    output = graph.get_tensor_by_name('...')

                    # create your custom loss
                    loss = custom_loss(output)

                    # add an optimizer of your choice
                    optimizer = tf.train.AdamOptimizer().minimize(loss)

                    return (loss, optimizer)

        loss (``string`` or ``None``):

            The name of the tensorflow tensor containing the loss, or ``None``
            if ``optimizer`` is a function.

        inputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of input tensors in the network to
            array keys.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output tensors in the network to array
            keys. New arrays will be generated by this node for each entry (if
            requested downstream).

        gradients (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output tensors in the network to
            array keys. New arrays containing the gradient of an output with
            respect to the loss will be generated by this node for each entry
            (if requested downstream).

        summary (``string`` or
                 ``dict``, ``string`` -> (``string`` (tensor name), freq),
                 optional):

            The name of the tensorflow tensor containing the tensorboard
            summaries or dictionary for different subcategories of summaires
            (key: string, value: tuple with tensor/op name and frequency,
            of evaluation).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        save_every (``int``, optional):

            After how many iterations to create a checkpoint to store the
            learnt weights.

        log_dir (``string``, optional):

            Directory for saving tensorboard summaries.

        log_every (``int``, optional):

            After how many iterations to write out tensorboard summaries.
    """

    def __init__(
        self,
        graph,
        optimizer,
        loss,
        inputs,
        outputs,
        gradients,
        summary=None,
        array_specs=None,
        save_every=2000,
        log_dir="./",
        log_every=1,
    ):
        super(Train, self).__init__(
            inputs, outputs, gradients, array_specs, spawn_subprocess=False
        )
        self.meta_graph_filename = graph
        self.optimizer_func = None
        self.optimizer_loss_names = None
        self.optimizer = None
        self.loss = None
        self.summary = summary
        self.session = None
        self.tf_gradient = {}
        self.graph = None
        self.basic_saver = None
        self.full_saver = None
        self.save_every = save_every
        self.iteration = None
        self.iteration_increment = None
        self.summary_saver = None
        self.log_dir = log_dir
        self.log_every = log_every
        # Check if optimizer is a str in python 2/3 compatible way.
        if isinstance(optimizer, ("".__class__, "".__class__)):
            self.optimizer_loss_names = (optimizer, loss)
        else:
            self.optimizer_func = optimizer

        # at least for some versions of tensorflow, the checkpoint name has to
        # start with a . if it is a relative path
        if not os.path.isabs(self.meta_graph_filename):
            self.meta_graph_filename = os.path.join(".", self.meta_graph_filename)

    def start(self):
        target = LocalServer.get_target()
        logger.info("Initializing tf session, connecting to %s...", target)

        self.graph = tf.Graph()
        self.session = tf.Session(target=target, graph=self.graph)

        with self.graph.as_default():
            self.__read_meta_graph()

        if self.summary is not None:
            self.summary_saver = tf.summary.FileWriter(self.log_dir, self.graph)

        if self.optimizer_func is None:
            # get actual operations/tensors from names
            self.optimizer = self.graph.get_operation_by_name(
                self.optimizer_loss_names[0]
            )
            self.loss = self.graph.get_tensor_by_name(self.optimizer_loss_names[1])

        # add symbolic gradients
        for tensor_name in self.gradients:
            tensor = self.graph.get_tensor_by_name(tensor_name)
            self.tf_gradient[tensor_name] = tf.gradients(self.loss, [tensor])[0]

    def train_step(self, batch, request):
        array_outputs = self.__collect_requested_outputs(request)
        inputs = self.__collect_provided_inputs(batch)

        to_compute = {
            "optimizer": self.optimizer,
            "loss": self.loss,
            "iteration": self.iteration_increment,
        }
        to_compute.update(array_outputs)

        # compute outputs, gradients, and update variables
        if isinstance(self.summary, str):
            to_compute["summaries"] = self.summary
        elif isinstance(self.summary, dict):
            for k, (v, f) in self.summary.items():
                if int(self.current_step + 1) % f == 0:
                    to_compute[k] = v
        outputs = self.session.run(to_compute, feed_dict=inputs)

        for array_key in array_outputs:
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(outputs[array_key], spec)

        batch.loss = outputs["loss"]
        batch.iteration = outputs["iteration"][0]
        self.current_step = batch.iteration
        if self.summary is not None:
            if isinstance(self.summary, str) and (
                batch.iteration % self.log_every == 0 or batch.iteration == 1
            ):
                self.summary_saver.add_summary(outputs["summaries"], batch.iteration)
            else:
                for k, (_, f) in self.summary.items():
                    if int(self.current_step) % f == 0:
                        self.summary_saver.add_summary(outputs[k], batch.iteration)

        if batch.iteration % self.save_every == 0:
            checkpoint_name = (
                self.meta_graph_filename + "_checkpoint_%i" % batch.iteration
            )

            logger.info("Creating checkpoint %s", checkpoint_name)

            self.full_saver.save(self.session, checkpoint_name)

    def stop(self):
        if self.session is not None:
            self.optimizer = None
            self.loss = None
            if self.summary is not None:
                self.summary_saver.close()
            self.session.close()
            self.graph = None
            self.session = None

    def __read_meta_graph(self):
        logger.info("Reading meta-graph...")

        # read the original meta-graph
        tf.train.import_meta_graph(
            self.meta_graph_filename + ".meta", clear_devices=True
        )

        # add custom gunpowder variables
        with tf.variable_scope("gunpowder"):
            self.iteration = tf.get_variable(
                "iteration", shape=1, initializer=tf.zeros_initializer, trainable=False
            )
            self.iteration_increment = tf.assign(self.iteration, self.iteration + 1)

        # Until now, only variables have been added to the graph that are part
        # of every checkpoint. We create a 'basic_saver' for only those
        # variables.
        self.basic_saver = tf.train.Saver(max_to_keep=None)

        # Add custom optimizer and loss, if requested. This potentially adds
        # more variables, not covered by the basic_saver.
        if self.optimizer_func is not None:
            loss, optimizer = self.optimizer_func(self.graph)
            self.loss = loss
            self.optimizer = optimizer

        # We create a 'full_saver' including those variables.
        self.full_saver = tf.train.Saver(max_to_keep=None)

        # find most recent checkpoint
        checkpoint_dir = os.path.dirname(self.meta_graph_filename)
        checkpoint = tf.train.latest_checkpoint(checkpoint_dir)

        if checkpoint:
            try:
                # Try to restore the graph, including the custom optimizer
                # state (if a custom optimizer was used).
                self.__restore_graph(checkpoint, restore_full=True)

            except tf.errors.NotFoundError:
                # If that failed, we just transitioned from an earlier training
                # without the custom optimizer. In this case, restore only the
                # variables of the original meta-graph and 'gunpowder'
                # variables. Custom optimizer variables will be default
                # initialized.
                logger.info("Checkpoint did not contain custom optimizer " "variables")
                self.__restore_graph(checkpoint, restore_full=False)
        else:
            logger.info("No checkpoint found")

            # initialize all variables
            self.session.run(tf.global_variables_initializer())

    def __restore_graph(self, checkpoint, restore_full):
        logger.info("Restoring model from %s", checkpoint)

        if restore_full:
            logger.info("...using a saver for all variables")
            self.full_saver.restore(self.session, checkpoint)

        else:
            # initialize all variables, such that non-basic variables are
            # initialized
            self.session.run(tf.global_variables_initializer())

            logger.info("...using a saver for basic variables only")
            self.basic_saver.restore(self.session, checkpoint)

        self.current_step = self.session.run(self.iteration)

    def __collect_requested_outputs(self, request):
        array_outputs = {}

        for output_name, array_key in self.outputs.items():
            if array_key in request:
                array_outputs[array_key] = output_name

        for output_name, array_key in self.gradients.items():
            if array_key in request:
                array_outputs[array_key] = self.tf_gradient[output_name]

        return array_outputs

    def __collect_provided_inputs(self, batch):
        inputs = {}

        for input_name, input_key in self.inputs.items():
            if isinstance(input_key, ArrayKey):
                if input_key in batch.arrays:
                    inputs[input_name] = batch.arrays[input_key].data
                else:
                    logger.warn(
                        "batch does not contain %s, input %s will not " "be set",
                        input_key,
                        input_name,
                    )
            elif isinstance(input_key, np.ndarray):
                inputs[input_name] = input_key
            elif isinstance(input_key, str):
                inputs[input_name] = getattr(batch, input_key)
            else:
                raise Exception(
                    "Unknown network input key {}, can't be given to "
                    "network".format(input_key)
                )

        return inputs



from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.batch_request import BatchRequest
from gunpowder.batch import Batch
from collections.abc import Iterable
import itertools
import logging
import numpy as np

logger = logging.getLogger(__name__)


class BalanceLabels(BatchFilter):
    """Creates a scale array to balance the loss between class labels.

    Note that this only balances loss weights per-batch and does not accumulate
    statistics about class balance across batches.

    Args:

        labels (:class:`ArrayKey`):

            An array containing binary or integer labels.

        scales (:class:`ArrayKey`):

            A array with scales to be created. This new array will have the
            same ROI and resolution as ``labels``.

        mask (:class:`ArrayKey`, optional):

            An optional mask (or list of masks) to consider for balancing.
            Every voxel marked with a 0 will not contribute to the scaling and
            will have a scale of 0 in ``scales``.

        slab (``tuple`` of ``int``, optional):

            A shape specification to perform the balancing in slabs of this
            size. -1 can be used to refer to the actual size of the label
            array. For example, a slab of::

                (2, -1, -1, -1)

            will perform the balancing for every each slice ``[0:2,:]``,
            ``[2:4,:]``, ... individually.

        num_classes(``int``, optional):

            The number of classes. Labels will be expected to be in the
            interval [0, ``num_classes``). Defaults to 2 for binary
            classification.

        clipmin (``float``, optional):

            Clip class fraction to clipmin when calculating class weights.
            Defaults to 0.05. Set to None if you do not want to clip min values.

        clipmax (``float``, optional):

            Clip class fraction to clipmax when calculating class weights.
            Defaults to 0.95. Set to None, if you do not want to clip max
            values.

    """

    def __init__(
        self,
        labels,
        scales,
        mask=None,
        slab=None,
        num_classes=2,
        clipmin=0.05,
        clipmax=0.95,
    ):
        self.labels = labels
        self.scales = scales
        if mask is None:
            self.masks = []
        elif not isinstance(mask, Iterable):
            self.masks = [mask]
        else:
            self.masks = mask

        self.slab = slab
        self.num_classes = num_classes
        self.clipmin = clipmin
        self.clipmax = clipmax

    def setup(self):
        assert self.labels in self.spec, (
            "Asked to balance labels %s, which are not provided." % self.labels
        )

        for mask in self.masks:
            assert mask in self.spec, (
                "Asked to apply mask %s to balance labels, but mask is not "
                "provided." % mask
            )

        spec = self.spec[self.labels].copy()
        spec.dtype = np.float32
        self.provides(self.scales, spec)
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.labels] = request[self.scales]
        for mask in self.masks:
            deps[mask] = request[self.scales]
        return deps

    def process(self, batch, request):
        labels = batch.arrays[self.labels]

        assert len(np.unique(labels.data)) <= self.num_classes, (
            "Found more unique labels than classes in %s." % self.labels
        )
        assert 0 <= np.min(labels.data) < self.num_classes, (
            "Labels %s are not in [0, num_classes)." % self.labels
        )
        assert 0 <= np.max(labels.data) < self.num_classes, (
            "Labels %s are not in [0, num_classes)." % self.labels
        )

        # initialize error scale with 1s
        error_scale = np.ones(labels.data.shape, dtype=np.float32)

        # set error_scale to 0 in masked-out areas
        for key in self.masks:
            mask = batch.arrays[key]
            assert (
                labels.data.shape == mask.data.shape
            ), "Shape of mask %s %s does not match %s %s" % (
                mask,
                mask.data.shape,
                self.labels,
                labels.data.shape,
            )
            error_scale *= mask.data

        if not self.slab:
            slab = error_scale.shape
        else:
            # slab with -1 replaced by shape
            slab = tuple(
                m if s == -1 else s for m, s in zip(error_scale.shape, self.slab)
            )

        slab_ranges = (range(0, m, s) for m, s in zip(error_scale.shape, slab))

        for start in itertools.product(*slab_ranges):
            slices = tuple(
                slice(start[d], start[d] + slab[d]) for d in range(len(slab))
            )
            self.__balance(labels.data[slices], error_scale[slices])

        spec = self.spec[self.scales].copy()
        spec.roi = labels.spec.roi

        outputs = Batch()
        outputs[self.scales] = Array(error_scale, spec)
        return outputs

    def __balance(self, labels, scale):
        labels = labels.astype(np.int64)

        # in the masked-in area, compute the fraction of per-class samples
        masked_in = scale.sum()
        classes, counts = np.unique(labels[np.nonzero(scale)], return_counts=True)
        fracs = (
            counts.astype(float) / masked_in if masked_in > 0 else np.zeros(counts.size)
        )
        if self.clipmin is not None or self.clipmax is not None:
            np.clip(fracs, self.clipmin, self.clipmax, fracs)

        # compute the class weights
        w_sparse = 1.0 / float(self.num_classes) / fracs
        w = np.zeros(self.num_classes)
        w[classes] = w_sparse

        # scale the masked-in scale with the class weights
        scale *= np.take(w, labels)



import logging

from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate

logger = logging.getLogger(__name__)


class Crop(BatchFilter):
    """Limits provided ROIs by either giving a new :class:`Roi` or crop
    fractions from either face of the provided ROI.

    Args:

        key (:class:`ArrayKey` or :class:`GraphKey`):

            The key of the array or points set to modify.

        roi (:class:`Roi` or ``None``):

            The ROI to crop to.

        fraction_negative (``tuple`` of ``float``):

            Relative crop starting from the negative end of the provided ROI.

        fraction_positive (``tuple`` of ``float``):

            Relative crop starting from the positive end of the provided ROI.
    """

    def __init__(self, key, roi=None, fraction_negative=None, fraction_positive=None):
        if roi is not None and (
            fraction_positive is not None or fraction_negative is not None
        ):
            raise RuntimeError(
                "'roi' and 'fraction_...' arguments can not be given together"
            )

        if (roi, fraction_positive, fraction_negative) == (None, None, None):
            raise RuntimeError("One of 'roi' and 'fraction_...' has to be given")

        if fraction_negative is not None and fraction_positive is None:
            fraction_positive = (0.0,) * len(fraction_negative)
        if fraction_positive is not None and fraction_negative is None:
            fraction_negative = (0.0,) * len(fraction_positive)

        self.key = key
        self.roi = roi
        self.fraction_negative = fraction_negative
        self.fraction_positive = fraction_positive

    def setup(self):
        spec = self.spec[self.key]

        if self.roi is not None:
            assert spec.roi.contains(
                self.roi
            ), "Crop ROI is not contained in upstream ROI."

            cropped_roi = self.roi

        else:
            total_fraction = tuple(
                n + p for n, p in zip(self.fraction_negative, self.fraction_positive)
            )
            if max(total_fraction) >= 1:
                raise RuntimeError("Sum of crop fractions exeeds 1")

            crop_positive = Coordinate(
                a * b for a, b in zip(spec.roi.shape, self.fraction_positive)
            )
            crop_negative = Coordinate(
                a * b for a, b in zip(spec.roi.shape, self.fraction_negative)
            )
            cropped_roi = spec.roi.grow(-crop_positive, -crop_negative)

        spec.roi = cropped_roi
        self.updates(self.key, spec)

    def process(self, batch, request):
        pass



from __future__ import print_function
import logging
import traceback

import sys
from typing import Optional, Any

logger = logging.getLogger(__name__)


class NoSuchModule(object):
    def __init__(self, name):
        self.__name = name
        self.__traceback_str = traceback.format_tb(sys.exc_info()[2])
        errtype, value = sys.exc_info()[:2]
        self.__exception = errtype(value)

    def __getattr__(self, item):
        raise self.__exception


try:
    import dvision
except ImportError:
    dvision = NoSuchModule("dvision")

try:
    import h5py
except ImportError:
    h5py = NoSuchModule("h5py")

try:
    import pyklb
except ImportError:
    pyklb = NoSuchModule("pyklb")

try:
    import tensorflow
except ImportError:
    tensorflow = NoSuchModule("tensorflow")

try:
    import torch
except ImportError:
    torch = NoSuchModule("torch")

try:
    import tensorboardX
except ImportError:
    tensorboardX = NoSuchModule("tensorboardX")

try:
    import malis
except ImportError:
    malis = NoSuchModule("malis")

try:
    import augment
except ImportError:
    augment = NoSuchModule("augment")

ZarrFile: Optional[Any] = None
try:
    import zarr
    from .zarr_file import ZarrFile
except ImportError:
    zarr = NoSuchModule("zarr")
    ZarrFile = None

try:
    import daisy
except ImportError:
    daisy = NoSuchModule("daisy")

try:
    import jax
except ImportError:
    jax = NoSuchModule("jax")

try:
    import jax.numpy as jnp
except ImportError:
    jnp = NoSuchModule("jnp")

try:
    import haiku
except ImportError:
    haiku = NoSuchModule("haiku")

try:
    import optax
except ImportError:
    optax = NoSuchModule("optax")



import logging
import numpy as np
from scipy.ndimage.morphology import distance_transform_edt

from .batch_filter import BatchFilter
from gunpowder.array import Array

logger = logging.getLogger(__name__)


class ExcludeLabels(BatchFilter):
    """Excludes several labels from the ground-truth.

    The labels will be replaced by background_value. An optional ignore mask
    will be created and set to 0 for the excluded locations that are further
    than a threshold away from not excluded locations.

    Args:

        labels (:class:`ArrayKey`):

            The array containing the labels.

        exclude (``list`` of ``int``):

            The labels to exclude from ``labels``.

        ignore_mask (:class:`ArrayKey`, optional):

            The ignore mask to create.

        ignore_mask_erode (``float``, optional):

            By how much (in world units) to erode the ignore mask.

        background_value (``int``, optional):

            Value to replace excluded IDs, defaults to 0.
    """

    def __init__(
        self, labels, exclude, ignore_mask=None, ignore_mask_erode=0, background_value=0
    ):
        self.labels = labels
        self.exclude = set(exclude)
        self.ignore_mask = ignore_mask
        self.ignore_mask_erode = ignore_mask_erode
        self.background_value = background_value

    def setup(self):
        assert (
            self.labels in self.spec
        ), "ExcludeLabels can only be used if GT_LABELS is provided upstream."
        if self.ignore_mask:
            self.provides(self.ignore_mask, self.spec[self.labels])

    def process(self, batch, request):
        gt = batch.arrays[self.labels]

        # 0 marks included regions (to be used directly with distance transform
        # later)
        include_mask = np.ones(gt.data.shape)

        gt_labels = np.unique(gt.data)
        logger.debug("batch contains GT labels: " + str(gt_labels))
        for label in gt_labels:
            if label in self.exclude:
                logger.debug("excluding label " + str(label))
                gt.data[gt.data == label] = self.background_value
            else:
                include_mask[gt.data == label] = 0

        # if no ignore mask is provided or requested, we are done
        if not self.ignore_mask or self.ignore_mask not in request:
            return

        voxel_size = self.spec[self.labels].voxel_size
        distance_to_include = distance_transform_edt(include_mask, sampling=voxel_size)
        logger.debug("max distance to foreground is " + str(distance_to_include.max()))

        # 1 marks included regions, plus a context area around them
        include_mask = distance_to_include < self.ignore_mask_erode

        # include mask was computed on labels ROI, we need to copy it to
        # the requested ignore_mask ROI
        gt_ignore_roi = request[self.ignore_mask].roi

        intersection = gt.spec.roi.intersect(gt_ignore_roi)
        intersection_in_gt = intersection - gt.spec.roi.offset
        intersection_in_gt_ignore = intersection - gt_ignore_roi.offset

        # to voxel coordinates
        intersection_in_gt //= voxel_size
        intersection_in_gt_ignore //= voxel_size

        gt_ignore = np.zeros((gt_ignore_roi // voxel_size).get_shape(), dtype=np.uint8)
        gt_ignore[intersection_in_gt_ignore.get_bounding_box()] = include_mask[
            intersection_in_gt.get_bounding_box()
        ]

        spec = self.spec[self.labels].copy()
        spec.roi = gt_ignore_roi
        spec.dtype = np.uint8
        batch.arrays[self.ignore_mask] = Array(gt_ignore, spec)



from copy import copy as shallow_copy
import logging
import multiprocessing

from .freezable import Freezable
from .profiling import ProfilingStats
from .array import Array, ArrayKey
from .graph import Graph, GraphKey

logger = logging.getLogger(__name__)


class Batch(Freezable):
    """Contains the requested batch as a collection of :class:`Arrays<Array>`
    and :class:`Graph` that is passed through the pipeline from sources to
    sinks.

    This collection mimics a dictionary. Items can be added with::

        batch = Batch()
        batch[array_key] = Array(...)
        batch[graph_key] = Graph(...)

    Here, ``array_key`` and ``graph_key`` are :class:`ArrayKey` and
    :class:`GraphKey`. The items can be queried with::

        array = batch[array_key]
        graph = batch[graph_key]

    Furthermore, pairs of keys/values can be iterated over using
    ``batch.items()``.

    To access only arrays or graphs, use the dictionaries ``batch.arrays``
    or ``batch.graphs``, respectively.

    Attributes:

        arrays (dict from :class:`ArrayKey` to :class:`Array`):

            Contains all arrays that have been requested for this batch.

        graphs (dict from :class:`GraphKey` to :class:`Graph`):

            Contains all graphs that have been requested for this batch.
    """

    __next_id = multiprocessing.Value("L")

    @staticmethod
    def get_next_id():
        with Batch.__next_id.get_lock():
            next_id = Batch.__next_id.value
            Batch.__next_id.value += 1
        return next_id

    def __init__(self):
        self.id = Batch.get_next_id()
        self.profiling_stats = ProfilingStats()
        self.arrays = {}
        self.graphs = {}
        self.affinity_neighborhood = None
        self.loss = None
        self.iteration = None

        self.freeze()

    def __setitem__(self, key, value):
        if isinstance(value, Array):
            assert isinstance(
                key, ArrayKey
            ), "Only a ArrayKey is allowed as key for an Array value."
            self.arrays[key] = value

        elif isinstance(value, Graph):
            assert isinstance(
                key, GraphKey
            ), "Only a GraphKey is allowed as key for Graph value."
            self.graphs[key] = value

        else:
            raise RuntimeError(
                "Only Array or Graph can be set in a %s." % type(self).__name__
            )

    def __getitem__(self, key):
        if isinstance(key, ArrayKey):
            return self.arrays[key]

        elif isinstance(key, GraphKey):
            return self.graphs[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey can be used as keys in a "
                "%s." % type(self).__name__
            )

    def __len__(self):
        return len(self.arrays) + len(self.graphs)

    def __contains__(self, key):
        if isinstance(key, ArrayKey):
            return key in self.arrays

        elif isinstance(key, GraphKey):
            return key in self.graphs

        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey can be used as keys in a "
                "%s. Key %s is a %s" % (type(self).__name__, key, type(key).__name__)
            )

    def __delitem__(self, key):
        if isinstance(key, ArrayKey):
            del self.arrays[key]

        elif isinstance(key, GraphKey):
            del self.graphs[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey can be used as keys in a "
                "%s." % type(self).__name__
            )

    def items(self):
        """Provides a generator iterating over key/value pairs."""

        for k, v in self.arrays.items():
            yield k, v
        for k, v in self.graphs.items():
            yield k, v

    def get_total_roi(self):
        """Get the union of all the array ROIs in the batch."""

        total_roi = None

        for _, array in self.arrays.items():
            if not array.spec.nonspatial:
                if total_roi is None:
                    total_roi = array.spec.roi
                else:
                    total_roi = total_roi.union(array.spec.roi)

        for _, graph in self.graphs.items():
            if total_roi is None:
                total_roi = graph.spec.roi
            else:
                total_roi = total_roi.union(graph.spec.roi)

        return total_roi

    def __repr__(self):
        r = "\n"
        for collection_type in [self.arrays, self.graphs]:
            for key, obj in collection_type.items():
                r += "\t%s: %s\n" % (key, obj.spec)
        return r

    def crop(self, request, copy=False):
        """Crop batch to meet the given request."""

        cropped = Batch()
        cropped.profiling_stats = self.profiling_stats
        cropped.loss = self.loss
        cropped.iteration = self.iteration

        for key, val in request.items():
            assert key in self, "%s not contained in this batch" % key
            if val.roi is None:
                cropped[key] = self[key]
            else:
                if isinstance(key, GraphKey):
                    cropped[key] = self[key].crop(val.roi)
                else:
                    cropped[key] = self[key].crop(val.roi, copy)

        return cropped

    def merge(self, batch, merge_profiling_stats=True):
        """Merge this batch (``a``) with another batch (``b``).

        This creates a new batch ``c`` containing arrays and graphs from
        both batches ``a`` and ``b``:

            * Arrays or Graphs that exist in either ``a`` or ``b`` will be
              referenced in ``c`` (not copied).

            * Arrays or Graphs that exist in both batches will keep only
              a reference to the version in ``b`` in ``c``.

        All other cases will lead to an exception.
        """

        merged = shallow_copy(self)

        for key, val in batch.items():
            # TODO: What is the goal of `val.spec.roi is None`? Why should that
            # mean that the key in merged gets overwritten?
            if key not in merged or val.spec.roi is None:
                merged[key] = val
            elif key in merged:
                merged[key] = val

        if merge_profiling_stats:
            merged.profiling_stats.merge_with(batch.profiling_stats)
        if batch.loss is not None:
            merged.loss = batch.loss
        if batch.iteration is not None:
            merged.iteration = batch.iteration

        return merged



import logging
import numpy as np

from gunpowder.batch_request import BatchRequest

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class Normalize(BatchFilter):
    """Normalize the values of an array to be floats between 0 and 1, based on
    the type of the array.

    Args:

        array (:class:`ArrayKey`):

            The key of the array to modify.

        factor (scalar, optional):

            The factor to use. If not given, a factor is chosen based on the
            ``dtype`` of the array (e.g., ``np.uint8`` would result in a factor
            of ``1.0/255``).

        dtype (data-type, optional):

            The datatype of the normalized array. Defaults to ``np.float32``.
    """

    def __init__(self, array, factor=None, dtype=np.float32):
        self.array = array
        self.factor = factor
        self.dtype = dtype

    def setup(self):
        self.enable_autoskip()
        array_spec = self.spec[self.array].copy()
        array_spec.dtype = self.dtype
        self.updates(self.array, array_spec)

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.array] = request[self.array]
        deps[self.array].dtype = None
        return deps

    def process(self, batch, request):
        if self.array not in batch.arrays:
            return

        factor = self.factor
        array = batch.arrays[self.array]
        array.spec.dtype = self.dtype

        if factor is None:
            logger.debug(
                "automatically normalizing %s with dtype=%s",
                self.array,
                array.data.dtype,
            )

            if array.data.dtype == np.uint8:
                factor = 1.0 / 255
            elif array.data.dtype == np.uint16:
                factor = 1.0 / (256 * 256 - 1)
            elif array.data.dtype == np.float32:
                assert array.data.min() >= 0 and array.data.max() <= 1, (
                    "Values are float but not in [0,1], I don't know how "
                    "to normalize. Please provide a factor."
                )
                factor = 1.0
            else:
                raise RuntimeError(
                    "Automatic normalization for "
                    + str(array.data.dtype)
                    + " not implemented, please "
                    "provide a factor."
                )

        logger.debug("scaling %s with %f", self.array, factor)
        array.data = array.data.astype(self.dtype) * factor



import distutils.util
import numpy as np
import logging
import requests
from copy import deepcopy

from gunpowder.batch import Batch
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.graph import GraphKey, Graph
from gunpowder.graph_spec import GraphSpec
from gunpowder.profiling import Timing

from gunpowder.graph import Node

logger = logging.getLogger(__name__)


class DvidPartnerAnnoationSourceReadException(Exception):
    pass


class MaskNotProvidedException(Exception):
    pass


# TODO: This seems broken. There is code involving a voxel size, but points
# don't have voxel sizes
class DvidPartnerAnnotationSource(BatchProvider):
    """
    :param hostname: hostname for DVID server
    :type hostname: str

    :param port: port for DVID server
    :type port: int

    :param uuid: UUID of node on DVID server
    :type uuid: str

    :param datasets: dict {GraphKey: DVID data instance}
    """

    def __init__(self, hostname, port, uuid, datasets=None, rois=None):
        self.hostname = hostname
        self.port = port
        self.url = "http://{}:{}".format(self.hostname, self.port)
        self.uuid = uuid

        self.datasets = datasets if datasets is not None else {}
        self.rois = rois if rois is not None else {}

        self.node_service = None
        self.dims = 0

    def setup(self):
        for points_key, points_name in self.datasets.items():
            self.provides(points_key, GraphSpec(roi=self.points_rois[points_key]))

        logger.info("DvidPartnerAnnotationSource.spec:\n{}".format(self.spec))

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        # if pre and postsynaptic locations requested, their id : SynapseLocation dictionaries should be created
        # together s.t. the ids are unique and allow to find partner locations
        if GraphKey.PRESYN in request.points or GraphKey.POSTSYN in request.points:
            try:  # either both have the same roi, or only one of them is requested
                assert (
                    request.points[GraphKey.PRESYN] == request.points[GraphKey.POSTSYN]
                )
            except AssertionError:
                assert (
                    GraphKey.PRESYN not in request.points
                    or GraphKey.POSTSYN not in request.points
                )
            if GraphKey.PRESYN in request.points:
                presyn_points, postsyn_points = self.__read_syn_points(
                    roi=request.points[GraphKey.PRESYN]
                )
            elif GraphKey.POSTSYN in request.points:
                presyn_points, postsyn_points = self.__read_syn_points(
                    roi=request.points[GraphKey.POSTSYN]
                )

        for points_key, roi in request.points.items():
            # check if requested points can be provided
            if points_key not in self.spec:
                raise RuntimeError(
                    "Asked for %s which this source does not provide" % points_key
                )
            # check if request roi lies within provided roi
            if not self.spec[points_key].roi.contains(roi):
                raise RuntimeError(
                    "%s's ROI %s outside of my ROI %s"
                    % (points_key, roi, self.spec[points_key].roi)
                )

            logger.debug("Reading %s in %s..." % (points_key, roi))
            id_to_point = {
                GraphKey.PRESYN: presyn_points,
                GraphKey.POSTSYN: postsyn_points,
            }[points_key]

            batch.points[points_key] = Graph(data=id_to_point, spec=GraphSpec(roi=roi))

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __load_json_annotations(
        self, array_shape_voxel, array_offset_voxel, array_name
    ):
        url = (
            "http://"
            + str(self.hostname)
            + ":"
            + str(self.port)
            + "/api/node/"
            + str(self.uuid)
            + "/"
            + str(array_name)
            + "/elements/{}_{}_{}/{}_{}_{}".format(
                array_shape_voxel[2],
                array_shape_voxel[1],
                array_shape_voxel[0],
                array_offset_voxel[2],
                array_offset_voxel[1],
                array_offset_voxel[0],
            )
        )
        annotations_file = requests.get(url)
        json_annotations = annotations_file.json()
        if json_annotations is None:
            json_annotations = []  # create empty_dummy_json_annotations
            # raise Exception ('No synapses found in region defined by array_offset {} and array_shape {}'.format(array_offset, array_shape))
        return json_annotations

    def __read_syn_points(self, roi):
        """read json file from dvid source, in json format to create for every location given"""

        if GraphKey.PRESYN in self.points_voxel_size:
            voxel_size = self.points_voxel_size[GraphKey.PRESYN]
        elif GraphKey.POSTSYN in self.points_voxel_size:
            voxel_size = self.points_voxel_size[GraphKey.POSTSYN]

        syn_file_json = self.__load_json_annotations(
            array_shape_voxel=roi.shape // voxel_size,
            array_offset_voxel=roi.offset // voxel_size,
            array_name=self.datasets[GraphKey.PRESYN],
        )

        presyn_points_dict, postsyn_points_dict = {}, {}
        location_to_location_id_dict, location_id_to_partner_locations = {}, {}
        for node_nr, node in enumerate(syn_file_json):
            # collect information
            kind = str(node["Kind"])
            location = (
                np.asarray((node["Pos"][2], node["Pos"][1], node["Pos"][0]))
                * voxel_size
            )
            location_id = int(node_nr)
            # some synapses are wrongly annotated in dvid source, have 'Tag': null ???, they are skipped
            try:
                syn_id = int(node["Tags"][0][3:])
            except:
                continue
            location_to_location_id_dict[str(location)] = location_id

            partner_locations = []
            try:
                for relation in node["Rels"]:
                    partner_locations.append(
                        (
                            np.asarray(
                                [
                                    relation["To"][2],
                                    relation["To"][1],
                                    relation["To"][0],
                                ]
                            )
                        )
                        * voxel_size
                    )
            except:
                partner_locations = []
            location_id_to_partner_locations[int(node_nr)] = partner_locations

            # check if property given, not always given
            props = {}
            if "conf" in node["Prop"]:
                props["conf"] = float(node["Prop"]["conf"])
            if "agent" in node["Prop"]:
                props["agent"] = str(node["Prop"]["agent"])
            if "flagged" in node["Prop"]:
                str_value_flagged = str(node["Prop"]["flagged"])
                props["flagged"] = bool(distutils.util.strtobool(str_value_flagged))
            if "multi" in node["Prop"]:
                str_value_multi = str(node["Prop"]["multi"])
                props["multi"] = bool(distutils.util.strtobool(str_value_multi))

            # create synPoint with information collected so far (partner_ids not completed yet)
            if kind == "PreSyn":
                syn_point = Node(
                    location=location,
                    location_id=location_id,
                    synapse_id=syn_id,
                    partner_ids=[],
                    props=props,
                )
                presyn_points_dict[int(node_nr)] = deepcopy(syn_point)
            elif kind == "PostSyn":
                syn_(
                    location=location,
                    location_id=location_id,
                    synapse_id=syn_id,
                    partner_ids=[],
                    props=props,
                )
                postsyn_points_dict[int(node_nr)] = deepcopy(syn_point)

        # add partner ids
        last_node_nr = len(syn_file_json) - 1
        for current_syn_point_id in location_id_to_partner_locations.keys():
            all_partner_ids = []
            for partner_loc in location_id_to_partner_locations[current_syn_point_id]:
                if location_to_location_id_dict.has_key(str(partner_loc)):
                    all_partner_ids.append(
                        int(location_to_location_id_dict[str(partner_loc)])
                    )
                else:
                    last_node_nr = last_node_nr + 1
                    assert not location_to_location_id_dict.has_key(str(partner_loc))
                    all_partner_ids.append(int(last_node_nr))

            if current_syn_point_id in presyn_points_dict:
                presyn_points_dict[current_syn_point_id].partner_ids = all_partner_ids
            elif current_syn_point_id in postsyn_points_dict:
                postsyn_points_dict[current_syn_point_id].partner_ids = all_partner_ids
            else:
                raise Exception("current syn_point id not found in any dictionary")

        return presyn_points_dict, postsyn_points_dict

    def __repr__(self):
        return "DvidPartnerAnnoationSource(hostname={}, port={}, uuid={}, raw_array_name={}, gt_array_name={}".format(
            self.hostname,
            self.port,
            self.uuid,
            self.array_names[ArrayKeys.RAW],
            self.array_names[ArrayKeys.GT_LABELS],
        )



from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest

from .batch_provider import BatchProvider

import random


class MergeProvider(BatchProvider):
    """Merges different providers::

        (a, b, c) + MergeProvider()

    will create a provider that combines the arrays and points offered by
    ``a``, ``b``, and ``c``. Array and point keys of ``a``, ``b``, and ``c`` should be
    the disjoint.
    """

    def __init__(self):
        self.key_to_provider = {}

    def setup(self):
        self.enable_placeholders()
        assert (
            len(self.get_upstream_providers()) > 0
        ), "at least one batch provider needs to be added to the MergeProvider"
        # Only allow merging if no two upstream_providers have the same
        # array/points keys
        error_message = (
            "Key {} provided by more than one upstream provider. Node MergeProvider only allows to merge "
            "providers with different keys."
        )
        for provider in self.get_upstream_providers():
            for key, spec in provider.spec.items():
                assert self.spec is None or key not in self.spec, error_message.format(
                    key
                )
                self.provides(key, spec)
                self.key_to_provider[key] = provider

    def provide(self, request):
        # create upstream requests
        upstream_requests = {}
        for key, spec in request.items():
            provider = self.key_to_provider[key]
            if provider not in upstream_requests:
                if request.is_deterministic():
                    random_seed = random.randint(0, 2**32)
                else:
                    random_seed = None
                upstream_requests[provider] = BatchRequest(random_seed=random_seed)

            upstream_requests[provider][key] = spec

        # execute requests, merge batches
        merged_batch = Batch()
        for provider, upstream_request in upstream_requests.items():
            batch = provider.request_batch(upstream_request)
            for key, array in batch.arrays.items():
                merged_batch.arrays[key] = array
            for key, graph in batch.graphs.items():
                merged_batch.graphs[key] = graph
            merged_batch.profiling_stats.merge_with(batch.profiling_stats)

        return merged_batch



from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.batch import Batch
from gunpowder.profiling import Timing
import numpy as np
import random


class Stack(BatchFilter):
    """Request several batches and stack them together, introducing a new
    dimension for each array. This is useful to create batches with several
    samples and only makes sense if there is a source of randomness upstream.

    This node stacks only arrays, not points. The resulting batch will have the
    same point sets as found in the first batch requested upstream.

    Args:

        num_repetitions (``int``):

            How many upstream batches to stack.
    """

    def __init__(self, num_repetitions):
        self.num_repetitions = num_repetitions

    def provide(self, request):
        batches = []
        for _ in range(self.num_repetitions):
            upstream_request = request.copy()
            if upstream_request.is_deterministic():
                # if the request is deterministic, create new seeds for each
                # upstream request (otherwise we would get the same batch over
                # and over). Using randint here is still deterministic, since
                # the RNG was already seeded with the requests original seed.
                seed = random.randint(0, 2**32)
                upstream_request._random_seed = seed
            batch = self.get_upstream_provider().request_batch(upstream_request)
            batches.append(batch)

        timing = Timing(self)
        timing.start()

        batch = Batch()
        for b in batches:
            batch.profiling_stats.merge_with(b.profiling_stats)

        for key, spec in request.array_specs.items():
            data = np.stack([b[key].data for b in batches])
            batch[key] = Array(data, batches[0][key].spec.copy())

        # copy points of first batch requested
        for key, spec in request.graph_specs.items():
            batch[key] = batches[0][key]

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch



import math
import logging
from random import random, randint, choices
import itertools

import numpy as np
from scipy.spatial import cKDTree
from skimage.transform import integral_image, integrate
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class RandomLocation(BatchFilter):
    """Choses a batch at a random location in the bounding box of the upstream
    provider.

    The random location is chosen such that the batch request ROI lies entirely
    inside the provider's ROI.

    If ``min_masked`` and ``mask`` are set, only batches are returned that have
    at least the given ratio of masked-in voxels. This is in general faster
    than using the :class:`Reject` node, at the expense of storing an integral
    array of the complete mask.

    If ``ensure_nonempty`` is set to a :class:`GraphKey`, only batches are
    returned that have at least one point of this point collection within the
    requested ROI.

    Additional tests for randomly picked locations can be implemented by
    subclassing and overwriting of :func:`accepts`. This method takes the
    randomly shifted request that meets all previous criteria (like
    ``min_masked`` and ``ensure_nonempty``) and should return ``True`` if the
    request is acceptable.

    Args:

        min_masked (``float``, optional):

            If non-zero, require that the random sample contains at least that
            ratio of masked-in voxels.

        mask (:class:`ArrayKey`, optional):

            The array to use for mask checks.

        ensure_nonempty (:class:`GraphKey`, optional):

            Ensures that when finding a random location, a request for
            ``ensure_nonempty`` will contain at least one point.

        p_nonempty (``float``, optional):

            If ``ensure_nonempty`` is set, it defines the probability that a
            request for ``ensure_nonempty`` will contain at least one point.
            Default value is 1.0.

        ensure_centered (``bool``, optional):

            if ``ensure_nonempty`` is set, ``ensure_centered`` guarantees
            that the center voxel of the roi contains a point.

        point_balance_radius (``int``):

            if ``ensure_nonempty`` is set, ``point_balance_radius`` defines
            a radius s.t. for every point `p` in ``ensure_nonempty``, the
            probability of picking p is inversely related to the number of
            other points within a distance of ``point_balance_radius`` to p.
            This helps avoid oversampling of dense regions of the graph, and
            undersampling of sparse regions.

        random_shift_key (``ArrayKey`` optional):

            if ``random_shift_key`` is not None, this node will populate
            that key with a nonspatial array containing the random shift
            used for each request. This can be useful for snapshot iterations
            if you want to figure out where that snapshot came from.
    """

    def __init__(
        self,
        min_masked=0,
        mask=None,
        ensure_nonempty=None,
        p_nonempty=1.0,
        ensure_centered=None,
        point_balance_radius=1,
        random_shift_key=None,
    ):
        self.min_masked = min_masked
        self.mask = mask
        self.mask_spec = None
        self.mask_integral = None
        self.ensure_nonempty = ensure_nonempty
        self.points = None
        self.p_nonempty = p_nonempty
        self.upstream_spec = None
        self.random_shift = None
        self.ensure_centered = ensure_centered
        self.point_balance_radius = point_balance_radius
        self.random_shift_key = random_shift_key

    def setup(self):
        upstream = self.get_upstream_provider()
        self.upstream_spec = upstream.spec

        if self.mask and self.min_masked > 0:
            assert self.mask in self.upstream_spec, (
                "Upstream provider does not have %s" % self.mask
            )
            self.mask_spec = self.upstream_spec.array_specs[self.mask]

            logger.info("requesting complete mask...")

            mask_request = BatchRequest({self.mask: self.mask_spec})
            mask_batch = upstream.request_batch(mask_request)

            logger.info("allocating mask integral array...")

            mask_data = mask_batch.arrays[self.mask].data
            mask_integral_dtype = np.uint64
            logger.debug("mask size is %s", mask_data.size)
            if mask_data.size < 2**32:
                mask_integral_dtype = np.uint32
            if mask_data.size < 2**16:
                mask_integral_dtype = np.uint16
            logger.debug("chose %s as integral array dtype", mask_integral_dtype)

            self.mask_integral = np.array(mask_data > 0, dtype=mask_integral_dtype)
            self.mask_integral = integral_image(self.mask_integral).astype(
                mask_integral_dtype
            )

        if self.ensure_nonempty:
            assert self.ensure_nonempty in self.upstream_spec, (
                "Upstream provider does not have %s" % self.ensure_nonempty
            )
            graph_spec = self.upstream_spec.graph_specs[self.ensure_nonempty]

            logger.info("requesting all %s points...", self.ensure_nonempty)

            nonempty_request = BatchRequest({self.ensure_nonempty: graph_spec})
            nonempty_batch = upstream.request_batch(nonempty_request)

            self.points = cKDTree(
                [p.location for p in nonempty_batch[self.ensure_nonempty].nodes]
            )

            point_counts = self.points.query_ball_point(
                [p.location for p in nonempty_batch[self.ensure_nonempty].nodes],
                r=self.point_balance_radius,
            )
            weights = [1 / len(point_count) for point_count in point_counts]
            self.cumulative_weights = list(itertools.accumulate(weights))

            logger.debug("retrieved %d points", len(self.points.data))

        # clear bounding boxes of all provided arrays and points --
        # RandomLocation does not have limits (offsets are ignored)
        for key, spec in self.spec.items():
            if spec.roi is not None:
                spec.roi.shape = Coordinate((None,) * spec.roi.dims)
                self.updates(key, spec)

        # provide randomness if asked for
        if self.random_shift_key is not None:
            self.provides(self.random_shift_key, ArraySpec(nonspatial=True))

    def prepare(self, request):
        logger.debug("request: %s", request.array_specs)
        logger.debug("my spec: %s", self.spec)

        if request.array_specs.keys():
            lcm_voxel_size = self.spec.get_lcm_voxel_size(request.array_specs.keys())
        else:
            lcm_voxel_size = Coordinate((1,) * request.get_total_roi().dims)

        shift_roi = self.__get_possible_shifts(request, lcm_voxel_size)

        if request.array_specs.keys():
            shift_roi = shift_roi.snap_to_grid(lcm_voxel_size, mode="shrink")
            lcm_shift_roi = shift_roi / lcm_voxel_size
            logger.debug(
                "restricting random locations to multiples of voxel size %s",
                lcm_voxel_size,
            )

        else:
            lcm_shift_roi = shift_roi

        assert not lcm_shift_roi.unbounded, (
            "Can not pick a random location, intersection of upstream ROIs is "
            "unbounded."
        )
        assert not lcm_shift_roi.empty, (
            "Can not satisfy batch request, no location covers all requested " "ROIs."
        )

        random_shift = self.__select_random_shift(
            request, lcm_shift_roi, lcm_voxel_size
        )

        self.random_shift = random_shift
        self.__shift_request(request, random_shift)

        return request

    def process(self, batch, request):
        if self.random_shift_key is not None:
            batch[self.random_shift_key] = Array(
                np.array(self.random_shift),
                ArraySpec(nonspatial=True),
            )

        # reset ROIs to request
        for array_key, spec in request.array_specs.items():
            batch.arrays[array_key].spec.roi = spec.roi
        for graph_key, spec in request.graph_specs.items():
            batch.graphs[graph_key].spec.roi = spec.roi

        # change shift point locations to lie within roi
        for graph_key in request.graph_specs.keys():
            batch.graphs[graph_key].shift(-self.random_shift)

    def accepts(self, request):
        """Should return True if the randomly chosen location is acceptable
        (besided meeting other criteria like ``min_masked`` and/or
        ``ensure_nonempty``). Subclasses can overwrite this method to implement
        additional tests for acceptable locations."""

        return True

    def __get_possible_shifts(self, request, voxel_size):
        total_shift_roi = None

        for key, spec in request.items():
            if spec.roi is None:
                continue

            request_roi = spec.roi
            provided_roi = self.upstream_spec[key].roi

            shift_roi = provided_roi.shift(-request_roi.begin).grow(
                (0,) * request_roi.dims, -(request_roi.shape - voxel_size)
            )

            if total_shift_roi is None:
                total_shift_roi = shift_roi
            else:
                if shift_roi != total_shift_roi:
                    total_shift_roi = total_shift_roi.intersect(shift_roi)

        logger.debug("valid shifts for request in " + str(total_shift_roi))

        return total_shift_roi

    def __select_random_shift(self, request, lcm_shift_roi, lcm_voxel_size):
        ensure_points = self.ensure_nonempty is not None and random() <= self.p_nonempty

        while True:
            if ensure_points:
                random_shift = self.__select_random_location_with_points(
                    request, lcm_shift_roi, lcm_voxel_size
                )
            else:
                random_shift = self.__select_random_location(
                    lcm_shift_roi, lcm_voxel_size
                )

            logger.debug("random shift: " + str(random_shift))

            if not self.__is_min_masked(random_shift, request):
                logger.debug("random location does not meet 'min_masked' criterium")
                continue

            if not self.__accepts(random_shift, request):
                logger.debug("random location does not meet user-provided criterium")
                continue

            return random_shift

    def __is_min_masked(self, random_shift, request):
        if not self.mask or self.min_masked == 0:
            return True

        # get randomly chosen mask ROI
        request_mask_roi = request.array_specs[self.mask].roi
        request_mask_roi = request_mask_roi.shift(random_shift)

        # get coordinates inside mask array
        mask_voxel_size = self.spec[self.mask].voxel_size
        request_mask_roi_in_array = request_mask_roi / mask_voxel_size
        request_mask_roi_in_array -= self.mask_spec.roi.offset / mask_voxel_size

        # get number of masked-in voxels
        num_masked_in = integrate(
            self.mask_integral,
            [request_mask_roi_in_array.begin],
            [
                request_mask_roi_in_array.end
                - Coordinate((1,) * self.mask_integral.ndim)
            ],
        )[0]

        mask_ratio = float(num_masked_in) / request_mask_roi_in_array.size
        logger.debug("mask ratio is %f", mask_ratio)

        return mask_ratio >= self.min_masked

    def __accepts(self, random_shift, request):
        # create a shifted copy of the request
        shifted_request = request.copy()
        self.__shift_request(shifted_request, random_shift)

        return self.accepts(shifted_request)

    def __shift_request(self, request, shift):
        # shift request ROIs
        for specs_type in [request.array_specs, request.graph_specs]:
            for key, spec in specs_type.items():
                if spec.roi is None:
                    continue
                roi = spec.roi.shift(shift)
                specs_type[key].roi = roi

    def __select_random_location_with_points(
        self, request, lcm_shift_roi, lcm_voxel_size
    ):
        request_points = request.graph_specs.get(self.ensure_nonempty)
        if request_points is None:
            total_roi = request.get_total_roi()
            logger.warning(
                f"Requesting non empty {self.ensure_nonempty}, however {self.ensure_nonempty} "
                f"has not been requested. Falling back on using the total roi of the "
                f"request {total_roi} for {self.ensure_nonempty}."
            )
            request_points_roi = total_roi
        else:
            request_points_roi = request_points.roi

        while True:
            # How to pick shifts that ensure that a randomly chosen point is
            # contained in the request ROI:
            #
            #
            # request          point
            # [---------)      .
            # 0        +10     17
            #
            #         least shifted to contain point
            #         [---------)
            #         8        +10
            #         ==
            #         point-request.begin-request.shape+1
            #
            #                  most shifted to contain point:
            #                  [---------)
            #                  17       +10
            #                  ==
            #                  point-request.begin
            #
            #         all possible shifts
            #         [---------)
            #         8        +10
            #         ==
            #         point-request.begin-request.shape+1
            #                   ==
            #                   request.shape

            # pick a random point
            point = choices(self.points.data, cum_weights=self.cumulative_weights)[0]

            logger.debug("select random point at %s", point)

            # get the lcm voxel that contains this point
            lcm_location = Coordinate(point / lcm_voxel_size)
            logger.debug("belongs to lcm voxel %s", lcm_location)

            # align the point request ROI with lcm voxel grid
            lcm_roi = request_points_roi.snap_to_grid(lcm_voxel_size, mode="shrink")
            lcm_roi = lcm_roi / lcm_voxel_size
            logger.debug("Point request ROI: %s", request_points_roi)
            logger.debug("Point request lcm ROI shape: %s", lcm_roi.shape)

            # get all possible starting points of lcm_roi.shape that contain
            # lcm_location
            if self.ensure_centered:
                lcm_shift_roi_begin = (
                    lcm_location
                    - lcm_roi.begin
                    - lcm_roi.shape / 2
                    + Coordinate((1,) * len(lcm_location))
                )
                lcm_shift_roi_shape = Coordinate((1,) * len(lcm_location))
            else:
                lcm_shift_roi_begin = (
                    lcm_location
                    - lcm_roi.begin
                    - lcm_roi.shape
                    + Coordinate((1,) * len(lcm_location))
                )
                lcm_shift_roi_shape = lcm_roi.shape
            lcm_point_shift_roi = Roi(lcm_shift_roi_begin, lcm_shift_roi_shape)

            logger.debug("lcm point shift roi: %s", lcm_point_shift_roi)

            # intersect with total shift ROI
            if not lcm_point_shift_roi.intersects(lcm_shift_roi):
                logger.debug(
                    "reject random shift, random point %s shift ROI %s does "
                    "not intersect total shift ROI %s",
                    point,
                    lcm_point_shift_roi,
                    lcm_shift_roi,
                )
                continue
            lcm_point_shift_roi = lcm_point_shift_roi.intersect(lcm_shift_roi)

            # select a random shift from all possible shifts
            random_shift = self.__select_random_location(
                lcm_point_shift_roi, lcm_voxel_size
            )
            logger.debug("random shift: %s", random_shift)

            # count all points inside the shifted ROI
            points = self.__get_points_in_roi(request_points_roi.shift(random_shift))
            assert (
                point in points
            ), "Requested batch to contain point %s, but got points " "%s" % (
                point,
                points,
            )
            num_points = len(points)

            return random_shift

    def __select_random_location(self, lcm_shift_roi, lcm_voxel_size):
        # select a random point inside ROI
        random_shift = Coordinate(
            randint(begin, end - 1)
            for begin, end in zip(lcm_shift_roi.begin, lcm_shift_roi.end)
        )

        random_shift *= lcm_voxel_size

        return random_shift

    def __get_points_in_roi(self, roi):
        points = []

        center = roi.center
        radius = math.ceil(float(max(roi.shape)) / 2)
        candidates = self.points.query_ball_point(center, radius, p=np.inf)

        for i in candidates:
            if roi.contains(self.points.data[i]):
                points.append(self.points.data[i])

        return np.array(points)



import numpy as np


def replace(array, old_values, new_values):
    """Replace all occurences of ``old_values[i]`` with ``new_values[i]`` in the
    given array."""

    old_values = np.array(old_values)
    new_values = np.array(new_values)

    values_map = np.arange(int(array.max() + 1), dtype=new_values.dtype)
    values_map[old_values] = new_values

    return values_map[array]



import logging
import random
import itertools

import numpy as np

from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate

logger = logging.getLogger(__name__)


class SimpleAugment(BatchFilter):
    """Randomly mirror and transpose all :class:`Arrays<Array>` and
    :class:`Graph` in a batch.

    Args:

        mirror_only (``list`` of ``int``, optional):

            If set, only mirror between the given axes. This is useful to
            exclude channels that have a set direction, like time.

        transpose_only (``list`` of ``int``, optional):

            If set, only transpose between the given axes. This is useful to
            limit the transpose to axes with the same resolution or to exclude
            non-spatial dimensions.

        mirror_probs (``list`` of ``float``, optional):

            If set, provides the probability for mirroring given axes. Default
            is 0.5 per axis. If given, must be given for every axis. i.e.
            [0,1,0] for 100% chance of mirroring axis 1 an no others.

        transpose_probs (``dict`` of ``tuple`` -> ``float``
        or ``list`` of ``float``, optional):

            The probability of transposing. If None, each transpose is equally
            likely.
            Can also be a dictionary of for ``tuple`` -> ``float``. For example
            {(0,1,2):0.5, (1,0,2):0.5} to define a 50% chance of transposing axes
            0 and 1. Note that if a provided option violates the `transpose_only`
            arg it will be dropped and remaining options will be reweighted.
            Can also be provided as a list of ``float``. i.e. [0.3, 0.5, 0.7].
            This will automatically generate a list of possible permutations
            and attempt to weight them appropriately. A weight of 0 means
            this axis will never be transposed, a weight of 1 means this axis
            will always be transposed.
    """

    def __init__(
        self,
        mirror_only=None,
        transpose_only=None,
        mirror_probs=None,
        transpose_probs=None,
    ):
        self.mirror_only = mirror_only
        self.mirror_probs = mirror_probs
        self.transpose_only = transpose_only
        self.transpose_probs = transpose_probs
        self.mirror_mask = None
        self.dims = None
        self.transpose_dims = None

    def setup(self):
        self.dims = self.spec.get_total_roi().dims

        # mirror_mask and transpose_dims refer to the indices of the spatial
        # dimensions only, starting counting at 0 for the first spatial
        # dimension

        if self.mirror_only is None:
            self.mirror_mask = [True] * self.dims
        else:
            self.mirror_mask = [d in self.mirror_only for d in range(self.dims)]
        if self.mirror_probs is None:
            self.mirror_probs = [0.5] * self.dims

        if self.transpose_only is None:
            self.transpose_dims = list(range(self.dims))
        else:
            self.transpose_dims = self.transpose_only
        if self.transpose_probs is None:
            self.permutation_dict = None
        elif isinstance(self.transpose_probs, list):
            self.permutation_dict = {}
            for permutation in itertools.permutations(range(self.dims), self.dims):
                total_prob = 1
                for i, j, p in zip(range(self.dims), permutation, self.transpose_probs):
                    if i not in self.transpose_dims and i != j:
                        total_prob = 0
                    else:
                        total_prob *= (1 - p) if i == j else p
                if total_prob > 0:
                    self.permutation_dict[permutation] = total_prob
        elif isinstance(self.transpose_probs, dict):
            self.permutation_dict = {}
            for k, v in self.transpose_probs.items():
                valid = True
                for i, j in enumerate(k):
                    if i not in self.transpose_only and i != j:
                        valid = False
                if valid:
                    self.permutation_dict[k] = v

    def prepare(self, request):
        self.mirror = [
            random.random() < self.mirror_probs[d] if self.mirror_mask[d] else 0
            for d in range(self.dims)
        ]

        if self.permutation_dict is not None:
            t = random.choices(
                list(self.permutation_dict.keys()),
                weights=list(self.permutation_dict.values()),
                k=1,
            )[0]
        else:
            t = random.sample(self.transpose_dims, k=len(self.transpose_dims))

        self.transpose = list(range(self.dims))
        for o, n in zip(self.transpose_dims, t):
            self.transpose[o] = n

        logger.debug("mirror = %s", self.mirror)
        logger.debug("transpose = %s", self.transpose)

        reverse_transpose = [0] * self.dims
        for d in range(self.dims):
            reverse_transpose[self.transpose[d]] = d

        logger.debug("downstream request = %s", request)

        self.__transpose_request(request, reverse_transpose)
        self.__mirror_request(request, self.mirror)

        logger.debug("upstream request = %s", request)

        return request

    def process(self, batch, request):
        # mirror and transpose ROIs of arrays & points in batch
        total_roi = batch.get_total_roi().copy()
        requested_keys = request.array_specs.keys()
        lcm_voxel_size = self.spec.get_lcm_voxel_size(requested_keys)

        for collection_type in [batch.arrays, batch.graphs]:
            for key, collector in collection_type.items():
                if key not in request:
                    continue
                if collector.spec.roi is None:
                    continue
                logger.debug("total ROI = %s", batch.get_total_roi())
                logger.debug("upstream %s ROI = %s", key, collector.spec.roi)
                self.__mirror_roi(collector.spec.roi, total_roi, self.mirror)
                logger.debug("mirrored %s ROI = %s", key, collector.spec.roi)
                self.__transpose_roi(
                    collector.spec.roi, total_roi, self.transpose, lcm_voxel_size
                )
                logger.debug("transposed %s ROI = %s", key, collector.spec.roi)

        mirror = tuple(slice(None, None, -1 if m else 1) for m in self.mirror)
        # arrays
        for array_key, array in batch.arrays.items():
            if array_key not in request:
                continue

            if array.spec.nonspatial:
                continue

            num_channels = len(array.data.shape) - self.dims
            channel_slices = (slice(None, None),) * num_channels

            array.data = array.data[channel_slices + mirror]

            transpose = [t + num_channels for t in self.transpose]
            array.data = array.data = array.data.transpose(
                list(range(num_channels)) + transpose
            )

        # graphs
        total_roi_offset = total_roi.offset
        total_roi_center = total_roi.center
        if lcm_voxel_size is not None:
            nearest_voxel_shift = Coordinate(
                (d % v) for d, v in zip(total_roi_center, lcm_voxel_size)
            )
            total_roi_center = total_roi_center - nearest_voxel_shift
        total_roi_end = total_roi.end
        logger.debug("augmenting in %s and center %s", total_roi, total_roi_center)

        for graph_key, graph in batch.graphs.items():
            if graph_key not in request:
                continue

            logger.debug("converting nodes in graph %s", graph_key)
            for node in list(graph.nodes):
                logger.debug("old location: %s, %s", node.id, node.location)

                # mirror
                location_in_total_offset = np.asarray(node.location) - total_roi_offset
                node.location = np.asarray(
                    [
                        total_roi_end[dim] - location_in_total_offset[dim]
                        if m
                        else node.location[dim]
                        for dim, m in enumerate(self.mirror)
                    ],
                    dtype=graph.spec.dtype,
                )

                logger.debug("after mirror: %s, %s", node.id, node.location)

                # transpose
                location_in_total_center = np.asarray(node.location) - total_roi_center

                if self.transpose != list(range(self.dims)):
                    for d in range(self.dims):
                        node.location[d] = (
                            location_in_total_center[self.transpose[d]]
                            + total_roi_center[d]
                        )

                logger.debug("after transpose: %s, %s", node.id, node.location)

                # due to the mirroring, points at the lower boundary of the ROI
                # could fall on the upper one, which excludes them from the ROI
                if not graph.spec.roi.contains(node.location):
                    graph.remove_node(node)

    def __mirror_request(self, request, mirror):
        total_roi = request.get_total_roi().copy()
        for key, spec in request.items():
            if spec.roi is not None:
                self.__mirror_roi(spec.roi, total_roi, mirror)

    def __transpose_request(self, request, transpose):
        total_roi = request.get_total_roi().copy()
        requested_keys = request.array_specs.keys()
        lcm_voxel_size = self.spec.get_lcm_voxel_size(requested_keys)
        for key, spec in request.items():
            if spec.roi is not None:
                self.__transpose_roi(spec.roi, total_roi, transpose, lcm_voxel_size)

    def __mirror_roi(self, roi, total_roi, mirror):
        total_roi_offset = total_roi.offset
        total_roi_shape = total_roi.shape

        roi_offset = roi.offset
        roi_shape = roi.shape

        roi_in_total_offset = roi_offset - total_roi_offset
        end_of_roi_in_total = roi_in_total_offset + roi_shape
        roi_in_total_offset_mirrored = total_roi_shape - end_of_roi_in_total
        roi_offset = Coordinate(
            total_roi_offset[d] + roi_in_total_offset_mirrored[d]
            if mirror[d]
            else roi_offset[d]
            for d in range(self.dims)
        )

        roi.offset = roi_offset

    def __transpose_roi(self, roi, total_roi, transpose, lcm_voxel_size):
        logger.debug("original roi = %s", roi)

        center = total_roi.center
        if lcm_voxel_size is not None:
            nearest_voxel_shift = Coordinate(
                (d % v) for d, v in zip(center, lcm_voxel_size)
            )
            center = center - nearest_voxel_shift
        logger.debug("center = %s", center)

        # Get distance from center, then transpose
        dist_to_center = center - roi.offset
        dist_to_center = Coordinate(
            dist_to_center[transpose[d]] for d in range(self.dims)
        )
        logger.debug("dist_to_center = %s", dist_to_center)

        # Using the tranposed distance to center, get the correct offset.
        new_offset = center - dist_to_center
        logger.debug("new_offset = %s", new_offset)

        shape = tuple(roi.shape[transpose[d]] for d in range(self.dims))
        roi.offset = new_offset
        roi.shape = shape
        logger.debug("tranposed roi = %s", roi)



from gunpowder.nodes import BatchProvider
from gunpowder.nodes.batch_provider import BatchRequestError

import logging
import traceback

logger = logging.getLogger(__name__)


class PipelineSetupError(Exception):
    def __init__(self, provider):
        self.provider = provider

    def __str__(self):
        return f"Exception in {self.provider.name()} while calling setup()"


class PipelineTeardownError(Exception):
    def __init__(self, provider):
        self.provider = provider

    def __str__(self):
        return f"Exception in {self.provider.name()} while calling teardown()"


class PipelineRequestError(Exception):
    def __init__(self, pipeline, request, original_traceback=None):
        self.pipeline = pipeline
        self.request = request
        self.original_traceback = original_traceback

    def __str__(self):
        return (
            (
                ("".join(self.original_traceback))
                if self.original_traceback is not None
                else ""
            )
            + "Exception in pipeline:\n"
            f"{self.pipeline}\n"
            "while trying to process request\n"
            f"{self.request}"
        )


class Pipeline:
    def __init__(self, node):
        """Create a pipeline from a single :class:`BatchProvider`."""

        assert isinstance(node, BatchProvider), f"{type(node)} is not a BatchProvider"

        self.output = node
        self.children = []
        self.initialized = False

    def traverse(self, callback, reverse=False):
        """Visit every node in the pipeline recursively (either from root to
        leaves of from leaves to the root if ``reverse`` is true). ``callback``
        will be called for each node encountered."""

        result = []

        if not reverse:
            result.append(callback(self))
        for child in self.children:
            result.append(child.traverse(callback, reverse))
        if reverse:
            result.append(callback(self))

        return result

    def copy(self):
        """Make a shallow copy of the pipeline."""

        pipeline = Pipeline(self.output)
        pipeline.children = [c.copy() for c in self.children]

        return pipeline

    def setup(self):
        """Connect all batch providers in the pipeline and call setup for
        each, from source to sink."""

        def connect(node):
            for child in node.children:
                node.output.add_upstream_provider(child.output)

        # connect all nodes
        self.traverse(connect)

        # call setup on all nodes
        if not self.initialized:

            def node_setup(node):
                try:
                    node.output.setup()
                except Exception as e:
                    raise PipelineSetupError(node.output) from e

            self.traverse(node_setup, reverse=True)
            self.initialized = True
        else:
            logger.warning(
                "pipeline.setup() called more than once (build() inside " "build()?)"
            )

    def internal_teardown(self):
        """Call teardown on each batch provider in the pipeline and disconnect
        all nodes."""

        try:

            def node_teardown(node):
                try:
                    node.output.internal_teardown()
                except Exception as e:
                    raise PipelineTeardownError(node.output) from e

            # call internal_teardown on all nodes
            self.traverse(node_teardown, reverse=True)
            self.initialized = False

        finally:
            # disconnect all nodes
            def disconnect(node):
                node.output.remove_upstream_providers()

            self.traverse(disconnect)

    def request_batch(self, request):
        """Request a batch from the pipeline."""

        try:
            return self.output.request_batch(request)
        except BatchRequestError as e:
            tb = traceback.format_exception(type(e), e, e.__traceback__)
            if isinstance(e, BatchRequestError):
                tb = tb[-1:]
            raise PipelineRequestError(self, request, original_traceback=tb) from None
        except Exception as e:
            raise PipelineRequestError(self, request) from e

    @property
    def spec(self):
        return self.output.spec

    def __add__(self, other):
        if isinstance(other, BatchProvider):
            other = Pipeline(other)

        if isinstance(other, Pipeline):
            result = other.copy()

            # add this pipeline as child to all leaves in other
            def add_self_to_leaves(node):
                if len(node.children) == 0:
                    node.children.append(self.copy())

            result.traverse(add_self_to_leaves, reverse=True)

        else:
            raise RuntimeError(f"Don't know how to add {type(other)} to Pipeline")

        return result

    def __radd__(self, other):
        assert isinstance(
            other, tuple
        ), f"Don't know how to radd {type(other)} to Pipeline"

        for o in other:
            assert isinstance(o, Pipeline) or isinstance(
                o, BatchProvider
            ), f"Don't know how to radd {type(o)} to Pipeline"

        other = tuple(
            Pipeline(o) if isinstance(o, BatchProvider) else o.copy() for o in other
        )

        result = self.copy()

        # add every other as child to leaves in pipeline
        def add_others_to_leaves(node):
            if len(node.children) == 0:
                for o in other:
                    node.children.append(o)

        result.traverse(add_others_to_leaves, reverse=True)

        return result

    def __repr__(self):
        def to_string(node):
            return node.output.name()

        reprs = self.traverse(to_string, reverse=True)

        return self.__rec_repr__(reprs)

    def __rec_repr__(self, reprs):
        if not isinstance(reprs, list):
            return str(reprs)

        num_children = len(reprs) - 1

        res = ""
        if num_children > 0:
            if num_children > 1:
                res = "("
            res += ", ".join(self.__rec_repr__(r) for r in reprs[:-1])
        if num_children > 0:
            if num_children > 1:
                res += ")"
            res += " -> "
        res += reprs[-1]
        return res



import math
from gunpowder.coordinate import Coordinate
from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.graph import GraphKey
from gunpowder.graph_spec import GraphSpec
from gunpowder.roi import Roi
from .freezable import Freezable
import logging
import copy

logger = logging.getLogger(__name__)


import logging

logger = logging.getLogger(__file__)


class ProviderSpec(Freezable):
    """A collection of (possibly partial) :class:`ArraySpecs<ArraySpec>` and
    :class:`GraphSpecs<GraphSpec>` describing a
    :class:`BatchProvider's<BatchProvider>` offered arrays and graphs.

    This collection mimics a dictionary. Specs can be added with::

        provider_spec = ProviderSpec()
        provider_spec[array_key] = ArraySpec(...)
        provider_spec[graph_key] = GraphSpec(...)

    Here, ``array_key`` and ``graph_key`` are :class:`ArrayKey` and
    :class:`GraphKey`. The specs can be queried with::

        array_spec = provider_spec[array_key]
        graph_spec = provider_spec[graph_key]

    Furthermore, pairs of keys/values can be iterated over using
    ``provider_spec.items()``.

    To access only array or graph specs, use the dictionaries
    ``provider_spec.array_specs`` or ``provider_spec.graph_specs``,
    respectively.

    Args:

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`):

            Initial array specs.

        graph_specs (``dict``, :class:`GraphKey` -> :class:`GraphSpec`):

            Initial graph specs.

    Attributes:

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`):

            Contains all array specs contained in this provider spec.

        graph_specs (``dict``, :class:`GraphKey` -> :class:`GraphSpec`):

            Contains all graph specs contained in this provider spec.
    """

    def __init__(self, array_specs=None, graph_specs=None):
        self.array_specs = {}
        self.graph_specs = {}
        self.freeze()

        # use __setitem__ instead of copying the dicts, this ensures type tests
        # are run
        if array_specs is not None:
            for key, spec in array_specs.items():
                self[key] = spec
        if graph_specs is not None:
            for key, spec in graph_specs.items():
                self[key] = spec

    def __setitem__(self, key, spec):
        assert isinstance(key, ArrayKey) or isinstance(key, GraphKey), (
            f"Only ArrayKey or GraphKey (not {type(key).__name__} are "
            "allowed as key for ProviderSpec, "
        )

        if isinstance(key, ArrayKey):
            if isinstance(spec, Roi):
                spec = ArraySpec(roi=spec)

            assert isinstance(spec, ArraySpec), (
                f"Only ArraySpec (not {type(spec).__name__}) can be set for " "ArrayKey"
            )

            self.array_specs[key] = spec.copy()

        else:
            if isinstance(spec, Roi):
                spec = GraphSpec(roi=spec)

            assert isinstance(spec, GraphSpec), (
                f"Only GraphSpec (not {type(spec).__name__}) can be set for " "GraphKey"
            )

            self.graph_specs[key] = spec.copy()

    def __getitem__(self, key):
        if isinstance(key, ArrayKey):
            return self.array_specs[key]

        elif isinstance(key, GraphKey):
            return self.graph_specs[key]
        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey can be used as keys in a "
                "%s." % type(self).__name__
            )

    def __len__(self):
        return len(self.array_specs) + len(self.graph_specs)

    def __contains__(self, key):
        if isinstance(key, ArrayKey):
            return key in self.array_specs

        elif isinstance(key, GraphKey):
            return key in self.graph_specs

        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey, can be used as keys in a "
                "%s. Key %s is a %s" % (type(self).__name__, key, type(key).__name__)
            )

    def __delitem__(self, key):
        if isinstance(key, ArrayKey):
            del self.array_specs[key]

        elif isinstance(key, GraphKey):
            del self.graph_specs[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or GraphKey can be used as keys in a "
                "%s." % type(self).__name__
            )

    def remove_placeholders(self):
        self.array_specs = {
            k: v for k, v in self.array_specs.items() if not v.placeholder
        }
        self.graph_specs = {
            k: v for k, v in self.graph_specs.items() if not v.placeholder
        }

    def items(self):
        """Provides a generator iterating over key/value pairs."""

        for k, v in self.array_specs.items():
            yield k, v
        for k, v in self.graph_specs.items():
            yield k, v

    def get_total_roi(self):
        """Get the union of all the ROIs."""

        total_roi = None
        for specs_type in [self.array_specs, self.graph_specs]:
            for _, spec in specs_type.items():
                if total_roi is None:
                    total_roi = spec.roi
                elif spec.roi is not None:
                    total_roi = total_roi.union(spec.roi)
        return total_roi

    def get_common_roi(self):
        """Get the intersection of all the requested ROIs."""

        common_roi = None
        for specs_type in [self.array_specs, self.graph_specs]:
            for _, spec in specs_type.items():
                if common_roi is None:
                    common_roi = spec.roi
                else:
                    common_roi = common_roi.intersect(spec.roi)

        return common_roi

    def get_lcm_voxel_size(self, array_keys=None):
        """Get the least common multiple of the voxel sizes in this spec.

        Args:

            array_keys (list of :class:`ArrayKey`, optional): If given,
                consider only the given array types.
        """

        if array_keys is None:
            array_keys = self.array_specs.keys()

        if not array_keys:
            return None

        lcm_voxel_size = None
        for key in array_keys:
            voxel_size = self.array_specs[key].voxel_size
            if voxel_size is None:
                continue
            if lcm_voxel_size is None:
                lcm_voxel_size = voxel_size
            else:
                lcm_voxel_size = Coordinate(
                    (
                        a * b // math.gcd(a, b)
                        for a, b in zip(lcm_voxel_size, voxel_size)
                    )
                )

        return lcm_voxel_size

    def copy(self):
        """Create a copy of this instance."""
        return copy.deepcopy(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            other_dict = copy.deepcopy(other.__dict__)
            self_dict = copy.deepcopy(self.__dict__)
            return self_dict == other_dict
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = "\n"
        for key, spec in self.items():
            r += "\t%s: %s\n" % (key, spec)
        return r



import logging
import numpy as np

from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.batch_request import BatchRequest
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate

logger = logging.getLogger(__name__)


def seg_to_affgraph(seg, nhood):
    nhood = np.array(nhood)

    # constructs an affinity graph from a segmentation
    # assume affinity graph is represented as:
    # shape = (e, z, y, x)
    # nhood.shape = (edges, 3)
    shape = seg.shape
    nEdge = nhood.shape[0]
    dims = nhood.shape[1]
    aff = np.zeros((nEdge,) + shape, dtype=np.int32)

    if dims == 2:
        for e in range(nEdge):
            aff[
                e,
                max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
            ] = (
                (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                    ]
                    == seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                    ]
                )
                * (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                    ]
                    > 0
                )
                * (
                    seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                    ]
                    > 0
                )
            )

    elif dims == 3:
        for e in range(nEdge):
            aff[
                e,
                max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
            ] = (
                (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                        max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
                    ]
                    == seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                        max(0, nhood[e, 2]) : min(shape[2], shape[2] + nhood[e, 2]),
                    ]
                )
                * (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                        max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
                    ]
                    > 0
                )
                * (
                    seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                        max(0, nhood[e, 2]) : min(shape[2], shape[2] + nhood[e, 2]),
                    ]
                    > 0
                )
            )

    else:
        raise RuntimeError(f"AddAffinities works only in 2 or 3 dimensions, not {dims}")

    return aff


class AddAffinities(BatchFilter):
    """Add an array with affinities for a given label array and neighborhood to
    the batch. Affinity values are created one for each voxel and entry in the
    neighborhood list, i.e., for each voxel and each neighbor of this voxel.
    Values are 1 iff both labels (of the voxel and the neighbor) are equal and
    non-zero.

    Args:

        affinity_neighborhood (``list`` of array-like):

            List of offsets for the affinities to consider for each voxel.

        labels (:class:`ArrayKey`):

            The array to read the labels from.

        affinities (:class:`ArrayKey`):

            The array to generate containing the affinities.

        labels_mask (:class:`ArrayKey`, optional):

            The array to use as a mask for ``labels``. Affinities connecting at
            least one masked out label will be masked out in
            ``affinities_mask``. If not given, ``affinities_mask`` will contain
            ones everywhere (if requested).

        unlabelled (:class:`ArrayKey`, optional):

            A binary array to indicate unlabelled areas with 0. Affinities from
            labelled to unlabelled voxels are set to 0, affinities between
            unlabelled voxels are masked out (they will not be used for
            training).

        affinities_mask (:class:`ArrayKey`, optional):

            The array to generate containing the affinitiy mask, as derived
            from parameter ``labels_mask``.
    """

    def __init__(
        self,
        affinity_neighborhood,
        labels,
        affinities,
        labels_mask=None,
        unlabelled=None,
        affinities_mask=None,
        dtype=np.uint8,
    ):
        self.affinity_neighborhood = np.array(affinity_neighborhood)
        self.labels = labels
        self.unlabelled = unlabelled
        self.labels_mask = labels_mask
        self.affinities = affinities
        self.affinities_mask = affinities_mask
        self.dtype = dtype

    def setup(self):
        assert self.labels in self.spec, (
            "Upstream does not provide %s needed by " "AddAffinities" % self.labels
        )

        voxel_size = self.spec[self.labels].voxel_size

        dims = self.affinity_neighborhood.shape[1]
        self.padding_neg = (
            Coordinate(
                min([0] + [a[d] for a in self.affinity_neighborhood])
                for d in range(dims)
            )
            * voxel_size
        )

        self.padding_pos = (
            Coordinate(
                max([0] + [a[d] for a in self.affinity_neighborhood])
                for d in range(dims)
            )
            * voxel_size
        )

        logger.debug("padding neg: " + str(self.padding_neg))
        logger.debug("padding pos: " + str(self.padding_pos))

        spec = self.spec[self.labels].copy()
        if spec.roi is not None:
            spec.roi = spec.roi.grow(self.padding_neg, -self.padding_pos)
        spec.dtype = self.dtype

        self.provides(self.affinities, spec)
        if self.affinities_mask:
            self.provides(self.affinities_mask, spec)
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()

        # grow labels ROI to accomodate padding
        labels_roi = request[self.affinities].roi.grow(
            -self.padding_neg, self.padding_pos
        )
        deps[self.labels] = request[self.affinities].copy()
        deps[self.labels].dtype = None
        deps[self.labels].roi = labels_roi

        if self.labels_mask:
            deps[self.labels_mask] = deps[self.labels].copy()
        if self.unlabelled:
            deps[self.unlabelled] = deps[self.labels].copy()

        return deps

    def process(self, batch, request):
        outputs = Batch()

        affinities_roi = request[self.affinities].roi

        logger.debug("computing ground-truth affinities from labels")

        affinities = seg_to_affgraph(
            batch.arrays[self.labels].data.astype(np.int32), self.affinity_neighborhood
        ).astype(self.dtype)

        # crop affinities to requested ROI
        offset = affinities_roi.offset
        shift = -offset - self.padding_neg
        crop_roi = affinities_roi.shift(shift)
        crop_roi /= self.spec[self.labels].voxel_size
        crop = crop_roi.get_bounding_box()

        logger.debug("cropping with " + str(crop))
        affinities = affinities[(slice(None),) + crop]

        spec = self.spec[self.affinities].copy()
        spec.roi = affinities_roi
        outputs.arrays[self.affinities] = Array(affinities, spec)

        if self.affinities_mask and self.affinities_mask in request:
            if self.labels_mask:
                logger.debug(
                    "computing ground-truth affinities mask from " "labels mask"
                )
                affinities_mask = seg_to_affgraph(
                    batch.arrays[self.labels_mask].data.astype(np.int32),
                    self.affinity_neighborhood,
                )
                affinities_mask = affinities_mask[(slice(None),) + crop]

            else:
                affinities_mask = np.ones_like(affinities)

            if self.unlabelled:
                # 1 for all affinities between unlabelled voxels
                unlabelled = 1 - batch.arrays[self.unlabelled].data
                unlabelled_mask = seg_to_affgraph(
                    unlabelled.astype(np.int32), self.affinity_neighborhood
                )
                unlabelled_mask = unlabelled_mask[(slice(None),) + crop]

                # 0 for all affinities between unlabelled voxels
                unlabelled_mask = 1 - unlabelled_mask

                # combine with mask
                affinities_mask = affinities_mask * unlabelled_mask

            affinities_mask = affinities_mask.astype(affinities.dtype)
            outputs.arrays[self.affinities_mask] = Array(affinities_mask, spec)

        else:
            if self.labels_mask is not None:
                logger.warning(
                    "GT labels does have a mask, but affinities "
                    "mask is not requested."
                )

        # Should probably have a better way of handling arbitrary batch attributes
        batch.affinity_neighborhood = self.affinity_neighborhood

        return outputs



import logging
import numpy as np
from scipy.ndimage.filters import gaussian_filter
from skimage import draw

from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.freezable import Freezable
from gunpowder.morphology import enlarge_binary_map, create_ball_kernel
from gunpowder.ndarray import replace
from gunpowder.graph_spec import GraphSpec
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)


class RasterizationSettings(Freezable):
    """Data structure to store parameters for rasterization of graph.

    Args:

        radius (``float`` or ``tuple`` of ``float``):

            The radius (for balls or tubes) or sigma (for peaks) in world units.

        mode (``string``):

            One of ``ball`` or ``peak``. If ``ball`` (the default), a ball with the
            given ``radius`` will be drawn. If ``peak``, the point will be
            rasterized as a peak with values :math:`\exp(-|x-p|^2/\sigma)` with
            sigma set by ``radius``.

        mask (:class:`ArrayKey`, optional):

            Used to mask the rasterization of points. The array is assumed to
            contain discrete labels. The object id at the specific point being
            rasterized is used to intersect the rasterization to keep it inside
            the specific object.

        inner_radius_fraction (``float``, optional):

            Only for mode ``ball``.

            If set, instead of a ball, a hollow sphere is rastered. The radius
            of the whole sphere corresponds to the radius specified with
            ``radius``. This parameter sets the radius of the hollow area, as a
            fraction of ``radius``.

        fg_value (``int``, optional):

            Only for mode ``ball``.

            The value to use to rasterize points, defaults to 1.

        bg_value (``int``, optional):

            Only for mode ``ball``.

            The value to use to for the background in the output array,
            defaults to 0.

        edges (``bool``, optional):

            Whether to rasterize edges by linearly interpolating between Nodes.
            Default is True.

        color_attr (``str``, optional)

            Which graph attribute to use for coloring nodes and edges. One
            useful example might be `component` which would color your graph
            based on the component labels.
            Notes:
            - Only available in "ball" mode
            - Nodes and Edges missing the attribute will be skipped.
            - color_attr must be populated for nodes and edges upstream of this node
    """

    def __init__(
        self,
        radius,
        mode="ball",
        mask=None,
        inner_radius_fraction=None,
        fg_value=1,
        bg_value=0,
        edges=True,
        color_attr=None,
    ):
        radius = np.array([radius]).flatten().astype(np.float64)

        if inner_radius_fraction is not None:
            assert (
                inner_radius_fraction > 0.0 and inner_radius_fraction < 1.0
            ), "Inner radius fraction has to be between (excluding) 0 and 1"
            inner_radius_fraction = 1.0 - inner_radius_fraction

        self.radius = radius
        self.mode = mode
        self.mask = mask
        self.inner_radius_fraction = inner_radius_fraction
        self.fg_value = fg_value
        self.bg_value = bg_value
        self.edges = edges
        self.color_attr = color_attr
        self.freeze()


class RasterizeGraph(BatchFilter):
    """Draw graphs into a binary array as balls/tubes of a given radius.

    Args:

        graph (:class:`GraphKey`):
            The key of the graph to rasterize.

        array (:class:`ArrayKey`):
            The key of the binary array to create.

        array_spec (:class:`ArraySpec`, optional):

            The spec of the array to create. Use this to set the datatype and
            voxel size.

        settings (:class:`RasterizationSettings`, optional):
            Which settings to use to rasterize the graph.
    """

    def __init__(self, graph, array, array_spec=None, settings=None):
        self.graph = graph
        self.array = array
        if array_spec is None:
            self.array_spec = ArraySpec()
        else:
            self.array_spec = array_spec
        if settings is None:
            self.settings = RasterizationSettings(1)
        else:
            self.settings = settings

    def setup(self):
        graph_roi = self.spec[self.graph].roi

        if self.array_spec.voxel_size is None:
            self.array_spec.voxel_size = Coordinate((1,) * graph_roi.dims)

        if self.array_spec.dtype is None:
            if self.settings.mode == "ball":
                self.array_spec.dtype = np.uint8
            else:
                self.array_spec.dtype = np.float32

        self.array_spec.roi = graph_roi.copy()
        self.provides(self.array, self.array_spec)

        self.enable_autoskip()

    def prepare(self, request):
        if self.settings.mode == "ball":
            context = np.ceil(self.settings.radius).astype(int)
        elif self.settings.mode == "peak":
            context = np.ceil(2 * self.settings.radius).astype(int)
        else:
            raise RuntimeError("unknown raster mode %s" % self.settings.mode)

        dims = self.array_spec.roi.dims
        if len(context) == 1:
            context = context.repeat(dims)

        # request graph in a larger area to get rasterization from outside
        # graph
        graph_roi = request[self.array].roi.grow(
            Coordinate(context), Coordinate(context)
        )

        # however, restrict the request to the graph actually provided
        graph_roi = graph_roi.intersect(self.spec[self.graph].roi)

        deps = BatchRequest()
        deps[self.graph] = GraphSpec(roi=graph_roi)

        if self.settings.mask is not None:
            mask_voxel_size = self.spec[self.settings.mask].voxel_size
            assert (
                self.spec[self.array].voxel_size == mask_voxel_size
            ), "Voxel size of mask and rasterized volume need to be equal"

            new_mask_roi = graph_roi.snap_to_grid(mask_voxel_size)
            deps[self.settings.mask] = ArraySpec(roi=new_mask_roi)

        return deps

    def process(self, batch, request):
        graph = batch.graphs[self.graph]
        mask = self.settings.mask
        voxel_size = self.spec[self.array].voxel_size

        # get roi used for creating the new array (graph_roi does not
        # necessarily align with voxel size)
        enlarged_vol_roi = graph.spec.roi.snap_to_grid(voxel_size)
        offset = enlarged_vol_roi.begin / voxel_size
        shape = enlarged_vol_roi.shape / voxel_size
        data_roi = Roi(offset, shape)

        logger.debug("Graph in %s", graph.spec.roi)
        for node in graph.nodes:
            logger.debug("%d, %s", node.id, node.location)
        logger.debug("Data roi in voxels: %s", data_roi)
        logger.debug("Data roi in world units: %s", data_roi * voxel_size)

        if graph.num_vertices == 0:
            # If there are no nodes at all, just create an empty matrix.
            rasterized_graph_data = np.zeros(
                data_roi.shape, dtype=self.spec[self.array].dtype
            )
        elif mask is not None:
            mask_array = batch.arrays[mask].crop(enlarged_vol_roi)
            # get those component labels in the mask, that contain graph
            labels = []
            # for i, point in graph.data.items():
            for i, point in enumerate(graph.nodes):
                v = Coordinate(point.location / voxel_size)
                v -= data_roi.begin
                labels.append(mask_array.data[v])
            # Make list unique
            labels = list(set(labels))

            # zero label should be ignored
            if 0 in labels:
                labels.remove(0)

            if len(labels) == 0:
                logger.debug(
                    "Graph and provided object mask do not overlap. No graph to rasterize."
                )
                rasterized_graph_data = np.zeros(
                    data_roi.shape, dtype=self.spec[self.array].dtype
                )
            else:
                # create data for the whole graph ROI, "or"ed together over
                # individual object masks
                rasterized_graph_data = np.sum(
                    [
                        self.__rasterize(
                            graph,
                            data_roi,
                            voxel_size,
                            self.spec[self.array].dtype,
                            self.settings,
                            Array(
                                data=(mask_array.data == label),
                                spec=mask_array.spec,
                            ),
                        )
                        for label in labels
                    ],
                    axis=0,
                    dtype=self.spec[self.array].dtype,
                )

        else:
            # create data for the whole graph ROI without mask
            rasterized_graph_data = self.__rasterize(
                graph, data_roi, voxel_size, self.spec[self.array].dtype, self.settings
            )

        # fix bg/fg labelling if requested
        if self.settings.bg_value != 0 or self.settings.fg_value != 1:
            replaced = replace(
                rasterized_graph_data,
                [0, 1],
                [self.settings.bg_value, self.settings.fg_value],
            )
            rasterized_graph_data = replaced.astype(self.spec[self.array].dtype)

        # create array and crop it to requested roi
        spec = self.spec[self.array].copy()
        spec.roi = data_roi * voxel_size
        rasterized_points = Array(data=rasterized_graph_data, spec=spec)
        batch[self.array] = rasterized_points.crop(request[self.array].roi)

    def __rasterize(
        self, graph, data_roi, voxel_size, dtype, settings, mask_array=None
    ):
        """Rasterize 'graph' into an array with the given 'voxel_size"""

        mask = mask_array.data if mask_array is not None else None

        logger.debug("Rasterizing graph in %s", graph.spec.roi)

        # prepare output array
        rasterized_graph = np.zeros(data_roi.shape, dtype=dtype)

        # Fast rasterization currently only implemented for mode ball without
        # inner radius set
        use_fast_rasterization = (
            settings.mode == "ball"
            and settings.inner_radius_fraction is None
            and len(list(graph.edges)) == 0
        )

        if use_fast_rasterization:
            dims = len(rasterized_graph.shape)

            # get structuring element for mode ball
            ball_kernel = create_ball_kernel(settings.radius, voxel_size)
            radius_voxel = Coordinate(np.array(ball_kernel.shape) / 2)
            data_roi_base = Roi(
                offset=Coordinate((0,) * dims), shape=Coordinate(rasterized_graph.shape)
            )
            kernel_roi_base = Roi(
                offset=Coordinate((0,) * dims), shape=Coordinate(ball_kernel.shape)
            )

        # Rasterize volume either with single voxel or with defined struct elememt
        for node in graph.nodes:
            # get the voxel coordinate, 'Coordinate' ensures integer
            v = Coordinate(node.location / voxel_size)

            # get the voxel coordinate relative to output array start
            v -= data_roi.begin

            # skip graph outside of mask
            if mask is not None and not mask[v]:
                continue

            logger.debug(
                "Rasterizing node %s at %s",
                node.location,
                node.location / voxel_size - data_roi.begin,
            )

            if use_fast_rasterization:
                # Calculate where to crop the kernel mask and the rasterized array
                shifted_kernel = kernel_roi_base.shift(v - radius_voxel)
                shifted_data = data_roi_base.shift(-(v - radius_voxel))
                arr_crop = data_roi_base.intersect(shifted_kernel)
                kernel_crop = kernel_roi_base.intersect(shifted_data)
                arr_crop_ind = arr_crop.get_bounding_box()
                kernel_crop_ind = kernel_crop.get_bounding_box()

                rasterized_graph[arr_crop_ind] = np.logical_or(
                    ball_kernel[kernel_crop_ind], rasterized_graph[arr_crop_ind]
                )

            else:
                if settings.color_attr is not None:
                    c = graph.nodes[node].get(settings.color_attr)
                    if c is None:
                        logger.debug(f"Skipping node: {node}")
                        continue
                    elif np.isclose(c, 1) and not np.isclose(settings.fg_value, 1):
                        logger.warning(
                            f"Node {node} is being colored with color {c} according to "
                            f"attribute {settings.color_attr} "
                            f"but color 1 will be replaced with fg_value: {settings.fg_value}"
                        )
                else:
                    c = 1
                rasterized_graph[v] = c
        if settings.edges:
            for e in graph.edges:
                if settings.color_attr is not None:
                    c = graph.edges[e].get(settings.color_attr)
                    if c is None:
                        continue
                    elif np.isclose(c, 1) and not np.isclose(settings.fg_value, 1):
                        logger.warning(
                            f"Edge {e} is being colored with color {c} according to "
                            f"attribute {settings.color_attr} "
                            f"but color 1 will be replaced with fg_value: {settings.fg_value}"
                        )

                u = graph.node(e.u)
                v = graph.node(e.v)
                u_coord = Coordinate(u.location / voxel_size)
                v_coord = Coordinate(v.location / voxel_size)
                line = draw.line_nd(u_coord, v_coord, endpoint=True)
                rasterized_graph[line] = 1

        # grow graph
        if not use_fast_rasterization:
            if settings.mode == "ball":
                enlarge_binary_map(
                    rasterized_graph,
                    settings.radius,
                    voxel_size,
                    settings.inner_radius_fraction,
                    in_place=True,
                )

            else:
                sigmas = settings.radius / voxel_size

                gaussian_filter(
                    rasterized_graph, sigmas, output=rasterized_graph, mode="constant"
                )

                # renormalize to have 1 be the highest value
                max_value = np.max(rasterized_graph)
                if max_value > 0:
                    rasterized_graph /= max_value

        if mask_array is not None:
            # use more efficient bitwise operation when possible
            if settings.mode == "ball":
                rasterized_graph &= mask
            else:
                rasterized_graph *= mask

        return rasterized_graph



import logging
import networkx as nx
import numpy as np

from gunpowder.batch import Batch
from gunpowder.graph import Graph
from gunpowder.graph_spec import GraphSpec
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.profiling import Timing

logger = logging.getLogger(__name__)


class GraphSource(BatchProvider):
    """Creates a gunpowder graph source from a daisy graph provider.
    Queries for graphs from a given Roi will only return edges completely
    contained within the Roi - edges that cross the boundary will not be
    included.

    Arguments:

        graph_provider (:class:`daisy.SharedGraphProvider`):
            A daisy graph provider to read the graph from.
            Can be backed by MongoDB or any other implemented backend.

        graph (:class:`GraphKey`):
            The key of the graph to create

        graph_spec (:class:`GraphSpec`, optional):
            An optional :class:`GraphSpec` containing a roi and optionally
            whether the graph is directed. The default is to have an unbounded
            roi and detect directedness from the graph_provider.
    """

    def __init__(self, graph_provider, graph, graph_spec=None):
        self.graph_provider = graph_provider
        self.graph = graph
        self.graph_spec = graph_spec

    def setup(self):
        if self.graph_spec is not None:
            roi = self.graph_spec.roi
            if self.graph_spec.directed is not None:
                assert self.graph_spec.directed == self.graph_provider.directed
        else:
            roi = None
        spec = GraphSpec(roi=roi, directed=self.graph_provider.directed)
        self.provides(self.graph, spec)

    def provide(self, request):
        timing = Timing(self)
        timing.start()
        batch = Batch()
        roi = request[self.graph].roi.copy()
        graph = GraphSource.create_gp_graph_from_daisy(self.graph_provider, roi)
        batch.graphs[self.graph] = graph

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    @staticmethod
    def create_gp_graph_from_daisy(graph_provider, roi):
        """A static method to convert a daisy graph into a gunpowder graph.
        Only includes edges if both endpoints are within the roi.

        Arguments:
            graph_provider (:class:`daisy.SharedGraphProvider`)
                A daisy graph provider to read the graph from

            roi (:class:`Roi`):
                The roi in which to read the graph

        Returns:
            An instance of :class:`Graph` containing the nodes and edges read
            from the daisy graph provider in the given roi.

        """
        logger.debug("Creating gunpowder graph from daisy graph provider")
        daisy_graph = graph_provider[roi]
        logger.debug("%d nodes found in roi %s", len(daisy_graph), roi)
        spec = GraphSpec(roi=roi, directed=daisy_graph.is_directed())
        dangling_nodes = []
        for node, data in daisy_graph.nodes(data=True):
            position_attribute = graph_provider.position_attribute
            if type(position_attribute) == list:
                if position_attribute[0] not in data:
                    dangling_nodes.append(node)
                    continue
                location = np.array(
                    [data[attr] for attr in position_attribute], dtype=np.float32
                )
            else:
                if position_attribute not in data:
                    dangling_nodes.append(node)
                    continue
                location = np.array(data[position_attribute], dtype=np.float32)
            data["location"] = location
            data["id"] = node

        logger.debug("Dangling nodes: %s", dangling_nodes)
        for n in dangling_nodes:
            daisy_graph.remove_node(n)

        if daisy_graph.is_directed():
            pure_nx_graph = nx.DiGraph()
        else:
            pure_nx_graph = nx.Graph()

        pure_nx_graph.update(daisy_graph)
        return Graph.from_nx_graph(pure_nx_graph, spec)



from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate
from gunpowder.array import ArrayKey, Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from gunpowder.batch import Batch
import logging
import numbers
import numpy as np

logger = logging.getLogger(__name__)


class UpSample(BatchFilter):
    """Upsample arrays in a batch by given factors.

    Args:

        source (:class:`ArrayKey`):

            The key of the array to upsample.

        factor (``int`` or ``Coordinate``):

            The factor to upsample with.

        target (:class:`ArrayKey`):

            The key of the array to store the upsampled ``source``.
    """

    def __init__(self, source, factor, target):
        assert isinstance(source, ArrayKey)
        assert isinstance(target, ArrayKey)
        assert isinstance(factor, numbers.Number) or isinstance(
            factor, Coordinate
        ), "Scaling factor should be a number or a Coordinate."

        self.source = source
        self.factor = factor
        self.target = target

    def setup(self):
        spec = self.spec[self.source].copy()

        if not isinstance(self.factor, Coordinate):
            self.factor = Coordinate((self.factor,) * spec.roi.dims)

        assert spec.voxel_size % self.factor == (0,) * len(
            spec.voxel_size
        ), "voxel size of upsampled volume is not integer: %s/%s = %s" % (
            spec.voxel_size,
            self.factor,
            tuple(v / f for v, f in zip(spec.voxel_size, self.factor)),
        )
        spec.voxel_size /= self.factor
        self.provides(self.target, spec)

    def prepare(self, request):
        deps = BatchRequest()

        if self.target not in request:
            return

        logger.debug("preparing upsampling of " + str(self.source))

        upstream_voxel_size = self.spec[self.source].voxel_size

        request_roi = request[self.target].roi.snap_to_grid(
            upstream_voxel_size, mode="grow"
        )
        logger.debug("request ROI is %s" % request_roi)

        # add or merge to batch request
        deps[self.source] = ArraySpec(roi=request_roi)

        return deps

    def process(self, batch, request):
        outputs = Batch()

        if self.target not in request:
            return

        input_roi = batch.arrays[self.source].spec.roi
        request_roi = request[self.target].roi

        assert input_roi.contains(request_roi)

        # upsample

        logger.debug("upsampling %s with %s", self.source, self.factor)

        crop = batch.arrays[self.source]
        data = crop.data

        for d, f in enumerate(self.factor):
            data = np.repeat(data, f, axis=-self.factor.dims + d)

        # create output array
        spec = self.spec[self.target].copy()
        spec.roi = input_roi
        outputs.arrays[self.target] = Array(data, spec).crop(request_roi)
        return outputs



import logging
import random

from .batch_filter import BatchFilter
from gunpowder.profiling import Timing

logger = logging.getLogger(__name__)


class Reject(BatchFilter):
    """Reject batches based on the masked-in vs. masked-out ratio.

    If a pipeline also contains a :class:`RandomLocation` node,
    :class:`Reject` needs to be placed downstream of it.

    Args:

        mask (:class:`ArrayKey`, optional):

            The mask to use, if any.

        min_masked (``float``, optional):

            The minimal required ratio of masked-in vs. masked-out voxels.
            Defaults to 0.5.

        ensure_nonempty (:class:`GraphKey`, optional)

            Ensures there is at least one point in the batch.

        reject_probability (``float``, optional):

            The probability by which a batch that is not valid (less than
            min_masked) is actually rejected. Defaults to 1., i.e. strict
            rejection.
    """

    def __init__(
        self, mask=None, min_masked=0.5, ensure_nonempty=None, reject_probability=1.0
    ):
        self.mask = mask
        self.min_masked = min_masked
        self.ensure_nonempty = ensure_nonempty
        self.reject_probability = reject_probability

    def setup(self):
        if self.mask:
            assert self.mask in self.spec, (
                "Reject can only be used if %s is provided" % self.mask
            )
        if self.ensure_nonempty:
            assert self.ensure_nonempty in self.spec, (
                "Reject can only be used if %s is provided" % self.ensure_nonempty
            )
        self.upstream_provider = self.get_upstream_provider()

    def provide(self, request):
        report_next_timeout = 10
        num_rejected = 0

        timing = Timing(self)
        timing.start()
        if self.mask:
            assert self.mask in request, (
                "Reject can only be used if %s is provided" % self.mask
            )
        if self.ensure_nonempty:
            assert self.ensure_nonempty in request, (
                "Reject can only be used if %s is provided" % self.ensure_nonempty
            )

        have_good_batch = False
        while not have_good_batch:
            batch = self.upstream_provider.request_batch(request)

            if self.mask:
                mask_ratio = batch.arrays[self.mask].data.mean()
            else:
                mask_ratio = None

            if self.ensure_nonempty:
                num_points = len(list(batch.graphs[self.ensure_nonempty].nodes))
            else:
                num_points = None

            have_min_mask = mask_ratio is None or mask_ratio > self.min_masked
            have_points = num_points is None or num_points > 0

            have_good_batch = have_min_mask and have_points

            if not have_good_batch and self.reject_probability < 1.0:
                have_good_batch = random.random() > self.reject_probability

            if not have_good_batch:
                if self.mask:
                    logger.debug(
                        "reject batch with mask ratio %f at %s",
                        mask_ratio,
                        batch.arrays[self.mask].spec.roi,
                    )
                if self.ensure_nonempty:
                    logger.debug(
                        "reject batch with empty points in %s",
                        batch.graphs[self.ensure_nonempty].spec.roi,
                    )
                num_rejected += 1

                if timing.elapsed() > report_next_timeout:
                    logger.warning(
                        "rejected %d batches, been waiting for a good one " "since %ds",
                        num_rejected,
                        report_next_timeout,
                    )
                    report_next_timeout *= 2

            else:
                if self.mask:
                    logger.debug(
                        "accepted batch with mask ratio %f at %s",
                        mask_ratio,
                        batch.arrays[self.mask].spec.roi,
                    )
                if self.ensure_nonempty:
                    logger.debug(
                        "accepted batch with nonempty points in %s",
                        self.ensure_nonempty,
                    )

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch



import numpy as np
from scipy.ndimage.morphology import distance_transform_edt


def enlarge_binary_map(
    binary_map, radius, voxel_size, ring_fraction=None, in_place=False
):
    """Enlarge existing regions in a binary map.

    Args:

        binary_map (numpy array):

            A matrix with zeros, in which regions to be enlarged are indicated
            with a 1 (regions can already represent larger areas).

        radius (``ndarray`` of ``float``):

            The amount by which to enlarge forground objects in world units.

        voxel_size (tuple, list or numpy array):

            Indicates the physical voxel size of the binary_map.

        ring_fraction (``float``, optional):

            If set, instead of just enlargin objects, a ring is grown around
            them (and the objects removed). The thickness of the ring is set
            with this parameter as a fraction of the radius.

        in_place (bool, optional):

            If set to ``True``, argument ``binary_map`` will be modified
            directly.

    Returns:

        A matrix with 0s and 1s of same dimension as input binary_map with
        enlarged regions (indicated with 1), unless ``in_place`` is set.
    """

    if len(np.unique(binary_map)) == 1:
        # Check whether there are regions at all. If there is no region (or
        # everything is full), return the same map.
        return binary_map

    if voxel_size is None:
        voxel_size = (1,) * binary_map.shape[0]

    voxel_size = np.asarray(voxel_size).astype(np.float32)

    # normalize, such that radius == 1 in all dimensions
    voxel_size = voxel_size / radius

    if in_place:
        np.logical_not(binary_map, out=binary_map)
    else:
        binary_map = np.logical_not(binary_map)
    edtmap = distance_transform_edt(binary_map, sampling=voxel_size)

    # grow objects
    if in_place:
        binary_map[:] = edtmap <= 1.0
    else:
        binary_map = edtmap <= 1.0

    # unmask inner part, if requested
    if ring_fraction is not None:
        binary_map[edtmap <= 1.0 - ring_fraction] = False

    if in_place:
        return None

    return binary_map


def create_ball_kernel(radius, voxel_size):
    """Generates a ball-shaped structuring element.

    Args:

        radius (``ndarray`` of ``float``):

            The radius of the ball-shaped structuring element in world-units.

        voxel_size (tuple, list or numpy array):

            Indicates the physical voxel size of the structuring element.

    Returns:

        The structuring element where elements of the neighborhood are 1 and 0
        otherwise. The shape of the returned array depends on radius and
        voxel_size. For instance voxel_size = [2, 1, 1], radius = 5 produces an
        array of shape (7, 11, 11)
    """
    voxel_size = np.asarray(voxel_size)

    # Calculate shape for new kernel, make it sufficiently large (--> ceil)
    radius_voxel = np.ceil(radius / voxel_size).astype(int)
    kernel_shape = np.array(radius_voxel) * 2 + 1

    kernel = np.zeros(kernel_shape, dtype=np.uint8)
    middle_point = kernel_shape // 2
    kernel[tuple(middle_point)] = 1

    enlarge_binary_map(kernel, radius, voxel_size, in_place=True)

    return kernel



import copy
import logging
import numpy as np

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import h5py
from gunpowder.profiling import Timing
from gunpowder.nodes.batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class Hdf5PointsSource(BatchProvider):
    """An HDF5 data source for :class:``Graph``. Currently only supports a
    specific case where graphs represent pre- and post-synaptic markers.

    Args:

        filename (string): The HDF5 file.

        datasets (dict): Dictionary of :class:``GraphKey`` -> dataset names
            that this source offers.

        rois (dict): Dictionary of :class:``GraphKey`` -> :class:``Roi`` to
            set the ROI for each point set provided by this source.

    """

    def __init__(self, filename, datasets, rois):
        self.filename = filename
        self.datasets = datasets
        self.rois = rois

        self.ndims = None

    def setup(self):
        hdf_file = h5py.File(self.filename, "r")

        for points_key, ds_name in self.datasets.items():
            if ds_name not in hdf_file:
                raise RuntimeError("%s not in %s" % (ds_name, self.filename))

            spec = PointsSpec()
            spec.roi = self.rois[points_key]

            self.provides(points_key, spec)

        hdf_file.close()

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        with h5py.File(self.filename, "r") as hdf_file:
            # if pre and postsynaptic locations required, their id
            # SynapseLocation dictionaries should be created together s.t. ids
            # are unique and allow to find partner locations
            if (
                PointsKeys.PRESYN in request.points_specs
                or PointsKeys.POSTSYN in request.points_specs
            ):
                assert (
                    request.points_specs[PointsKeys.PRESYN].roi
                    == request.points_specs[PointsKeys.POSTSYN].roi
                )
                # Cremi specific, ROI offset corresponds to offset present in the
                # synapse location relative to the raw data.
                dataset_offset = self.spec[PointsKeys.PRESYN].roi.offset
                presyn_points, postsyn_points = self.__get_syn_points(
                    roi=request.points_specs[PointsKeys.PRESYN].roi,
                    syn_file=hdf_file,
                    dataset_offset=dataset_offset,
                )

            for points_key, request_spec in request.points_specs.items():
                logger.debug("Reading %s in %s...", points_key, request_spec.roi)
                id_to_point = {
                    PointsKeys.PRESYN: presyn_points,
                    PointsKeys.POSTSYN: postsyn_points,
                }[points_key]

                points_spec = self.spec[points_key].copy()
                points_spec.roi = request_spec.roi
                batch.points[points_key] = Points(data=id_to_point, spec=points_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __get_syn_points(self, roi, syn_file, dataset_offset=None):
        presyn_points_dict, postsyn_points_dict = {}, {}
        presyn_node_ids = syn_file["annotations/presynaptic_site/partners"][
            :, 0
        ].tolist()
        postsyn_node_ids = syn_file["annotations/presynaptic_site/partners"][
            :, 1
        ].tolist()

        for node_nr, node_id in enumerate(syn_file["annotations/ids"]):
            location = syn_file["annotations/locations"][node_nr]
            if dataset_offset is not None:
                logging.debug(
                    "adding global offset to points %i %i %i"
                    % (dataset_offset[0], dataset_offset[1], dataset_offset[2])
                )
                location += dataset_offset

            # cremi synapse locations are in physical space
            if roi.contains(Coordinate(location)):
                if node_id in presyn_node_ids:
                    kind = "PreSyn"
                    assert syn_file["annotations/types"][node_nr] == "presynaptic_site"
                    syn_id = int(np.where(presyn_node_ids == node_id)[0])
                    partner_node_id = postsyn_node_ids[syn_id]
                elif node_id in postsyn_node_ids:
                    kind = "PostSyn"
                    assert syn_file["annotations/types"][node_nr] == "postsynaptic_site"
                    syn_id = int(np.where(postsyn_node_ids == node_id)[0])
                    partner_node_id = presyn_node_ids[syn_id]
                else:
                    raise Exception("Node id neither pre- no post-synaptic")

                partners_ids = [int(partner_node_id)]
                location_id = int(node_id)

                props = {}
                if node_id in syn_file["annotations/comments/target_ids"]:
                    props = {"unsure": True}

                # create synpaseLocation & add to dict
                if kind == "PreSyn":
                    syn_point = PreSynPoint(
                        location=location,
                        location_id=location_id,
                        synapse_id=syn_id,
                        partner_ids=partners_ids,
                        props=props,
                    )
                    presyn_points_dict[int(node_id)] = copy.deepcopy(syn_point)
                elif kind == "PostSyn":
                    syn_point = PostSynPoint(
                        location=location,
                        location_id=location_id,
                        synapse_id=syn_id,
                        partner_ids=partners_ids,
                        props=props,
                    )
                    postsyn_points_dict[int(node_id)] = copy.deepcopy(syn_point)

        return presyn_points_dict, postsyn_points_dict

    def __repr__(self):
        return self.filename



try:
    import Queue
except:
    import queue as Queue
import logging
import multiprocessing
import os
import sys

import numpy as np

logger = logging.getLogger(__name__)


class NoResult(Exception):
    pass


class ParentDied(Exception):
    pass


class WorkersDied(Exception):
    pass


class ProducerPool(object):
    def __init__(self, callables, queue_size=10):
        self.__watch_dog = multiprocessing.Process(
            target=self._run_watch_dog, args=(callables,)
        )
        self.__stop = multiprocessing.Event()
        self.__result_queue = multiprocessing.Queue(queue_size)

    def __del__(self):
        self.stop()

    def start(self):
        """Start the pool of producers."""

        if self.__watch_dog is None:
            raise RuntimeError("can't start a ProducerPool a second time")

        if self.__watch_dog.is_alive():
            logger.warning("trying to start workers, but they are already running")
            return

        self.__stop.clear()
        self.__watch_dog.start()

    def get(self, timeout=0):
        """Return the next result from the producer pool.

        If timeout is set and there is not result after the given number of
        seconds, exception NoResult is raised.
        """

        block = False
        if timeout == 0:
            timeout = 1
            block = True

        item = None
        while item == None:
            try:
                item = self.__result_queue.get(timeout=timeout)
            except Queue.Empty:
                if not block:
                    raise NoResult()

        if isinstance(item, Exception):
            raise item
        return item

    def stop(self):
        """Stop the pool of producers.

        Items currently being produced will not be waited for and be discarded."""

        if self.__watch_dog is None:
            return

        self.__stop.set()
        if self.__watch_dog._popen is not None:
            self.__watch_dog.join()
        self.__watch_dog = None

    def _run_watch_dog(self, callables):
        parent_pid = os.getppid()

        logger.debug("watchdog started with PID " + str(os.getpid()))
        logger.debug("parent PID " + str(parent_pid))

        workers = [
            multiprocessing.Process(target=self._run_worker, args=(c,))
            for c in callables
        ]

        try:
            logger.debug("starting %d workers" % len(workers))
            for worker in workers:
                worker.start()

            while not self.__stop.wait(1):
                if os.getppid() != parent_pid:
                    logger.error("parent of producer pool died, shutting down")
                    self.__result_queue.put(ParentDied())
                    break
                if not self._all_workers_alive(workers):
                    logger.error("at least one of my workers died, shutting down")
                    self.__result_queue.put(WorkersDied())
                    break
        except:
            pass

        finally:
            logger.info("terminating workers...")
            for worker in workers:
                worker.terminate()

            logger.info("joining workers...")
            for worker in workers:
                worker.join()

            logger.info("done")

    def _run_worker(self, target):
        parent_pid = os.getppid()

        logger.debug("worker started with PID " + str(os.getpid()))
        logger.debug("parent PID " + str(parent_pid))

        result = None
        np.random.seed(None)
        while True:
            if os.getppid() != parent_pid:
                logger.debug("worker %d: watch-dog died, stopping" % os.getpid())
                break

            if result is None:
                try:
                    result = target()
                except Exception as e:
                    result = e
                    # don't stop on normal exceptions -- place them in result queue
                    # and let them be handled by caller
                except:
                    logger.error("received error: " + str(sys.exc_info()[0]))
                    # this is most likely a keyboard interrupt, stop process
                    break

            try:
                self.__result_queue.put(result, timeout=1)
                result = None
            except Queue.Full:
                logger.debug(
                    "worker %d: result queue is full, waiting to place my result"
                    % os.getpid()
                )

        logger.debug("worker with PID " + str(os.getpid()) + " exiting")
        os._exit(1)

    def _all_workers_alive(self, workers):
        return all([worker.is_alive() for worker in workers])



from gunpowder.array import ArrayKey, Array
from gunpowder.array_spec import ArraySpec
from gunpowder.ext import jax
from gunpowder.nodes.generic_predict import GenericPredict
from gunpowder.jax import GenericJaxModel

import pickle
import logging
from typing import Dict, Union, Optional

logger = logging.getLogger(__name__)


class Predict(GenericPredict):
    """JAX implementation of :class:`gunpowder.nodes.Predict`.

    Args:

        model (subclass of ``gunpowder.jax.GenericJaxModel``):

            The model to use for prediction.

        inputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of input tensors in the network to
            array keys.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output tensors in the network to array
            keys. New arrays will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        checkpoint: (``string``, optional):

            An optional path to the saved parameters for your jax module.
            These will be loaded and used for prediction if provided.

        spawn_subprocess (bool, optional): Whether to run ``predict`` in a
            separate process. Default is false.
    """

    def __init__(
        self,
        model: GenericJaxModel,
        inputs: Dict[str, ArrayKey],
        outputs: Dict[Union[str, int], ArrayKey],
        array_specs: Optional[Dict[ArrayKey, ArraySpec]] = None,
        checkpoint: Optional[str] = None,
        spawn_subprocess=False,
    ):
        self.array_specs = array_specs if array_specs is not None else {}

        super(Predict, self).__init__(
            inputs, outputs, array_specs, spawn_subprocess=spawn_subprocess
        )

        self.model = model
        self.checkpoint = checkpoint
        self.model_params = None

    def start(self):
        if self.checkpoint is not None:
            with open(self.checkpoint, "rb") as f:
                self.model_params = pickle.load(f)

    def predict(self, batch, request):
        inputs = self.get_inputs(batch)

        if self.model_params is None:
            # need to init model first
            rng = jax.random.PRNGKey(request.random_seed)
            self.model_params = self.model.initialize(rng, inputs)

        out = jax.jit(self.model.forward)(self.model_params, inputs)
        outputs = self.get_outputs(out, request)
        self.update_batch(batch, request, outputs)

    def get_inputs(self, batch):
        model_inputs = {
            key: jax.device_put(batch[value].data) for key, value in self.inputs.items()
        }
        return model_inputs

    def get_outputs(self, module_out, request):
        outputs = {}
        for key, value in self.outputs.items():
            if value in request:
                outputs[value] = module_out[key]
        return outputs

    def update_batch(self, batch, request, requested_outputs):
        for array_key, tensor in requested_outputs.items():
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(tensor, spec)

    def stop(self):
        pass



import numpy as np
import skimage

from gunpowder.batch_request import BatchRequest

from .batch_filter import BatchFilter


class NoiseAugment(BatchFilter):
    """Add random noise to an array. Uses the scikit-image function skimage.util.random_noise.
    See scikit-image documentation for more information on arguments and additional kwargs.

    Args:

        array (:class:`ArrayKey`):

            The intensity array to modify. Should be of type float and within range [-1, 1] or [0, 1].

        mode (``string``):

            Type of noise to add, see scikit-image documentation.

        clip (``bool``):

            Whether to preserve the image range (either [-1, 1] or [0, 1]) by clipping values in the end, see
            scikit-image documentation
    """

    def __init__(self, array, mode="gaussian", clip=True, **kwargs):
        self.array = array
        self.mode = mode
        self.clip = clip
        self.kwargs = kwargs

    def setup(self):
        self.enable_autoskip()
        self.updates(self.array, self.spec[self.array])

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.array] = request[self.array].copy()
        return deps

    def process(self, batch, request):
        raw = batch.arrays[self.array]

        assert raw.data.dtype == np.float32 or raw.data.dtype == np.float64, (
            "Noise augmentation requires float types for the raw array (not "
            + str(raw.data.dtype)
            + "). Consider using Normalize before."
        )
        if self.clip:
            assert (
                raw.data.min() >= -1 and raw.data.max() <= 1
            ), "Noise augmentation expects raw values in [-1,1] or [0,1]. Consider using Normalize before."

        seed = request.random_seed

        try:
            raw.data = skimage.util.random_noise(
                raw.data, mode=self.mode, rng=seed, clip=self.clip, **self.kwargs
            ).astype(raw.data.dtype)

        except ValueError:
            # legacy version of skimage random_noise
            raw.data = skimage.util.random_noise(
                raw.data, mode=self.mode, seed=seed, clip=self.clip, **self.kwargs
            ).astype(raw.data.dtype)



import logging
import itertools
import numpy as np

from gunpowder.array import Array
from gunpowder.nodes.batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class AddBlobsFromPoints(BatchFilter):
    """Add an array with blobs at locations given by a specified points
    collection. The blobs are also restricted to stay within the same class in
    the restrictive_mask array that corresponds to the center voxel of the
    blob.

    Args:

        blob_settings(dict):

            Where each desired output blob map should have it's own entry
            consisting of the following format:

            `blob_name` : dict (

                'points_key' : Desired point type to use for blob locations

                'output_array_key': Desired array type name for output map

                'output_array_dtype': Desired array dtype name for output map

                'radius': Desired radius of blobs, since blobs are restricted
                by the restricting mask, this radius should be thought of as
                the maximum radius of the blobs.

                'output_voxel_size': voxel_size of output array. Voxel size of
                restrictive mask

                'restrictive_mask_key': Array type of restrictive mask

                'id_mapper': Functor (class with a __call__ function) that can
                take an ID and map it to some other value. This class should
                also have a 'make_map' method that will be called at the
                beggining of each process step and given all ids in all arrays
                to be processed for that batch.
            )

            This is an example blob_setting for presynaptic blobs in the cremi
            dataset::

              add_blob_data = {
                'PRESYN': {
                  'points_key': PointsTypes.PRESYN,
                  'output_array_key': ArrayTypes.PRESYN_BLOB,
                  'output_array_dtype': 'int64',
                  'radius': 60,
                  'output_voxel_size': voxel_size,
                  'restrictive_mask_key': ArrayTypes.GT_LABELS,
                  'max_desired_overlap': 0.05
                }
              }
    """

    def __init__(self, blob_settings):
        self.blob_settings = blob_settings

        for points_key, settings in self.blob_settings.items():
            blob_settings[points_key]["blob_placer"] = BlobPlacer(
                radius=settings["radius"],
                voxel_size=settings["output_voxel_size"],
                dtype=settings["output_array_dtype"],
            )

    def setup(self):
        for blob_name, settings in self.blob_settings.items():
            self.provides(
                settings["output_array_key"],
                self.spec[settings["restrictive_mask_key"]],
            )

    def prepare(self, request):
        for blob_name, settings in self.blob_settings.items():
            array_key = settings["output_array_key"]
            if array_key in request:
                points_key = settings["points_key"]
                request_roi = request[array_key].roi

                # If point is not already requested, add to request
                if points_key not in request.points_specs:
                    request.add(points_key, request_roi.shape)
                else:
                    request[points_key].roi = request[points_key].roi.union(request_roi)

                # Get correct size for restrictive_mask_key
                restrictive_mask_key = settings["restrictive_mask_key"]
                if restrictive_mask_key not in request.array_specs:
                    request.add(restrictive_mask_key, request_roi.shape)
                else:
                    request[restrictive_mask_key].roi = request[
                        restrictive_mask_key
                    ].roi.union(request_roi)
            else:
                # do nothing if no blobs of this type were requested
                logger.warning(
                    "%s output array type for %s never requested. \
                    Deleting entry..."
                    % (settings["output_array_key"], blob_name)
                )
                del self.blob_settings[blob_name]

    def process(self, batch, request):
        # check arrays and gather all IDs and synapse IDs
        all_points = {}
        all_synapse_ids = {}

        for blob_name, settings in self.blob_settings.items():
            # Unpack settings
            points_key = settings["points_key"]
            restrictive_mask_key = settings["restrictive_mask_key"]

            # Make sure both the necesary point types and arrays are present
            assert points_key in batch.points, (
                "Upstream does not provide required point type\
            : %s"
                % points_key
            )

            assert restrictive_mask_key in batch.arrays, (
                "Upstream does not provide required \
            array type: %s"
                % restrictive_mask_key
            )

            # Get point data
            points = batch.points[points_key]

            # If point doesn't have it's corresponding partner, delete it
            if (
                "partner_points" in settings.keys()
                and settings["partner_points"] is not None
            ):
                partner_points = batch.points[settings["partner_points"]]
                synapse_ids = []
                for point_id, point in points.data.items():
                    # pdb.set_trace()
                    if point.partner_ids[0] not in partner_points.data.keys():
                        logger.warning(
                            "Point %s has no partner. Deleting..." % point_id
                        )
                        del points.data[point_id]
                    else:
                        synapse_ids.append(point.synapse_id)

            all_synapse_ids[points_key] = synapse_ids
            all_points[points_key] = points

        for blob_name, settings in self.blob_settings.items():
            # Unpack settings
            points_key = settings["points_key"]
            array_key = settings["output_array_key"]
            voxel_size = settings["output_voxel_size"]
            restrictive_mask_key = settings["restrictive_mask_key"]
            restrictive_mask = batch.arrays[restrictive_mask_key].crop(
                request[array_key].roi
            )

            id_mapper = settings["id_mapper"]
            dtype = settings["output_array_dtype"]

            if id_mapper is not None:
                id_mapper.make_map(all_points)

            # Initialize output array
            shape_array = np.asarray(request[array_key].roi.shape) / voxel_size
            blob_map = np.zeros(shape_array, dtype=dtype)

            # Get point data
            points = batch.points[points_key]

            offset = np.asarray(points.spec.roi.offset)
            for point_id, point_data in points.data.items():
                voxel_location = np.round(
                    ((point_data.location - offset) / (voxel_size))
                ).astype("int32")

                synapse_id = point_data.synapse_id
                # if mapping exists, do it
                if id_mapper is not None:
                    synapse_id = id_mapper(synapse_id)

                settings["blob_placer"].place(
                    blob_map, voxel_location, synapse_id, restrictive_mask.data
                )

            # Provide array
            batch.arrays[array_key] = Array(blob_map, spec=request[array_key].copy())
            batch.arrays[array_key].spec.dtype = dtype

            # add id_mapping to attributes
            if id_mapper is not None:
                id_map_list = np.array(list(id_mapper.get_map().items()))
                batch.arrays[array_key].attrs["id_mapping"] = id_map_list

            batch.arrays[array_key].attrs["point_ids"] = points.data.keys()
            batch.arrays[array_key].attrs["synapse_ids"] = all_synapse_ids[points_key]

            # Crop all other requests
        for array_key, array in request.array_specs.items():
            batch.arrays[array_key] = batch.arrays[array_key].crop(array.roi)

        for points_key, points in request.points_specs.items():
            batch.points[points_key] = batch.points[points_key].spec.roi = points.roi


class BlobPlacer:
    """Places synapse array blobs from location data.
    Args:
        radius: int - that desired radius of synaptic blobs
        voxel_size: array, list, tuple - voxel size in physical
    """

    def __init__(self, radius, voxel_size, dtype="uint64"):
        self.voxel_size = voxel_size
        if isinstance(self.voxel_size, (list, tuple)):
            self.voxel_size = np.asarray(self.voxel_size)

        self.radius = radius / self.voxel_size
        self.sphere_map = np.zeros(self.radius * 2, dtype=dtype)
        self.center = (np.asarray(self.sphere_map.shape)) / 2

        ranges = [
            range(0, self.radius[0] * 2),
            range(0, self.radius[1] * 2),
            range(0, self.radius[2] * 2),
        ]

        for index in np.asarray(list(itertools.product(*ranges))):
            # if distance less than r, place a 1
            if np.linalg.norm((self.center - index) * self.voxel_size) <= radius:
                self.sphere_map[tuple(index)] = 1

        self.sphere_voxel_array = np.sum(self.sphere_map, axis=(0, 1, 2))

    def place(self, matrix, location, marker, mask):
        """Places synapse
        Args:
            matrix: 4D np array - 1st dim are for layers to avoid overlap
            (3 should be more than enough)
            location: np array - location where to place synaptic blob within given matrix
            marker: int - the ID used to mark this paricular synapse in the matrix
            mask:   3D np array - when placing a blob, will sample mask at
        center location and only place blob in interection where mask has
        the same ID. Usually used to restrict synaptic blobs inside their
        respective cells (using segmentation)
        """
        # Calculate cube circumscribing the sphere to place
        start = location - self.radius
        end = location + self.radius

        # check if sphere fits in matrix
        if np.all(start >= 0) and np.all(np.asarray(matrix.shape) - end >= 0):
            # calculate actual synapse shape from intersection between sphere and restrictive mask
            restricting_label = mask[location[0], location[1], location[2]]

            restricting_mask = (
                mask[start[0] : end[0], start[1] : end[1], start[2] : end[2]]
                == restricting_label
            )

            shape = self.sphere_map * restricting_mask

            # place shape in chosen layer
            matrix[start[0] : end[0], start[1] : end[1], start[2] : end[2]] += (
                shape * marker
            )
            return matrix, True

        logger.warning("Location %s out of bounds" % (location))
        return matrix, False



class Freezable(object):
    __isfrozen = False

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError("%r is frozen, you can't add attributes to it" % self)
        object.__setattr__(self, key, value)

    def freeze(self):
        self.__isfrozen = True

    def thaw(self):
        self.__isfrozen = False



from gunpowder.ext import h5py
from .hdf5like_source_base import Hdf5LikeSource


class Hdf5Source(Hdf5LikeSource):
    """An HDF5 data source.

    Provides arrays from HDF5 datasets. If the attribute ``resolution`` is set
    in a HDF5 dataset, it will be used as the array's ``voxel_size``. If the
    attribute ``offset`` is set in a dataset, it will be used as the offset of
    the :class:`Roi` for this array. It is assumed that the offset is given in
    world units.

    Args:

        filename (``string``):

            The HDF5 file.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary of array keys to dataset names that this source offers.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to array specs to overwrite
            the array specs automatically determined from the data file. This
            is useful to set a missing ``voxel_size``, for example. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.

        channels_first (``bool``, optional):

            Specifies the ordering of the dimensions of the HDF5-like data source.
            If channels_first is set (default), then the input shape is expected
            to be (channels, spatial dimensions). This is recommended because of
            better performance. If channels_first is set to false, then the input
            data is read in channels_last manner and converted to channels_first.
    """

    def _open_file(self, filename):
        return h5py.File(filename, "r")
